import sys 
import os 
from PySide6 .QtWidgets import QApplication ,QMainWindow ,QWidget ,QHBoxLayout ,QSplitter ,QVBoxLayout ,QLabel ,QTabWidget ,QComboBox ,QGroupBox ,QCheckBox ,QPushButton ,QFileDialog ,QGridLayout ,QTextEdit ,QScrollArea ,QMessageBox ,QSizePolicy ,QSpinBox ,QStyleFactory 
from PySide6 .QtCore import Qt 
from PySide6 .QtGui import QTextOption ,QPalette ,QColor 
from matplotlib .backends .backend_qt5agg import FigureCanvasQTAgg as FigureCanvas 
from matplotlib .backends .backend_qt5agg import NavigationToolbar2QT as NavigationToolbar 
from matplotlib .figure import Figure 
import numpy as np 
import json 
from sklearn .ensemble import RandomForestClassifier ,GradientBoostingClassifier ,AdaBoostClassifier ,ExtraTreesClassifier 
from sklearn .model_selection import KFold 
from sklearn .metrics import confusion_matrix ,accuracy_score ,f1_score ,recall_score ,precision_score 
from lightgbm import LGBMClassifier 
from sklearn .neighbors import KNeighborsClassifier 
from sklearn .neural_network import MLPClassifier 
from sklearn .naive_bayes import GaussianNB 
from sklearn .tree import DecisionTreeClassifier 
from sklearn .linear_model import LogisticRegression 
from sklearn .preprocessing import LabelEncoder 
from sklearn .svm import SVC 
import torch 
import torch .nn as nn 
import torch .optim as optim 
from torch .utils .data import DataLoader ,TensorDataset 
import platform 
import pickle 
import tensorflow as tf 
from datetime import datetime 
if torch .cuda .is_available ():
    import cudf 
    import cuml 



def save_classifiers (classifiers ,save_path ):
    for classifier_name ,classifier_data in classifiers .items ():
        model =classifier_data ['model']

        if isinstance (model ,(tf .keras .Sequential ,tf .keras .Model )):

            file_name =f"{classifier_name}.h5"
            file_path =os .path .join (save_path ,file_name )
            model .save (file_path )
        elif classifier_name .startswith ("RAPIDS"):

            file_name =f"{classifier_name}.pkl"
            file_path =os .path .join (save_path ,file_name )
            with open (file_path ,'wb')as file :
                pickle .dump (model ,file )
        else :

            file_name =f"{classifier_name}.pkl"
            file_path =os .path .join (save_path ,file_name )
            with open (file_path ,'wb')as file :
                pickle .dump (model ,file )

def load_classifiers (classifier_files ):
    classifiers ={}
    for file_path in classifier_files :
        classifier_name =os .path .splitext (os .path .basename (file_path ))[0 ]

        if file_path .endswith (".h5"):

            model =tf .keras .models .load_model (file_path )
        elif file_path .endswith (".pkl"):

            with open (file_path ,'rb')as file :
                model =pickle .load (file )

        classifiers [classifier_name ]=model 

    return classifiers 

def load_data (npz_file_path ):
    with np .load (npz_file_path ,allow_pickle =True )as data :
        X =data ['X']
        settings =json .loads (data ['settings'].item ())
        label =settings .get ('ML_tag')
    return X ,label 

def prepare_data (npz_file_paths ):
    features =[]
    labels =[]
    min_events =float ('inf')

    for file_path in npz_file_paths :
        X ,label =load_data (file_path )
        min_events =min (min_events ,len (X ))

    for file_path in npz_file_paths :
        X ,label =load_data (file_path )
        if len (X )>min_events :
            indices =np .random .choice (len (X ),min_events ,replace =False )
            X =X [indices ]

        for event_features in X :
            event_features =np .nan_to_num (event_features ,nan =0.0 ,posinf =0.0 ,neginf =0.0 )
            features .append (event_features )
            labels .append (label )

    return np .array (features ),np .array (labels )


class Net (nn .Module ):
    def __init__ (self ,input_size ,hidden_size ,num_classes ):
        super (Net ,self ).__init__ ()
        self .fc1 =nn .Linear (input_size ,hidden_size )
        self .relu =nn .ReLU ()
        self .fc2 =nn .Linear (hidden_size ,num_classes )

    def forward (self ,x ):
        out =self .fc1 (x )
        out =self .relu (out )
        out =self .fc2 (out )
        return out 

def train_pytorch_model (X_train ,y_train ,input_size ,hidden_size ,num_classes ,num_epochs ,batch_size ,learning_rate ):
    label_encoder =LabelEncoder ()
    y_train_encoded =label_encoder .fit_transform (y_train )

    train_dataset =TensorDataset (torch .from_numpy (X_train ),torch .from_numpy (y_train_encoded ))
    train_loader =DataLoader (train_dataset ,batch_size =batch_size ,shuffle =True )

    model =Net (input_size ,hidden_size ,num_classes )

    criterion =nn .CrossEntropyLoss ()
    optimizer =optim .Adam (model .parameters (),lr =learning_rate )

    for epoch in range (num_epochs ):
        for batch_idx ,(data ,target )in enumerate (train_loader ):
            optimizer .zero_grad ()
            output =model (data .float ())
            loss =criterion (output ,target .long ())
            loss .backward ()
            optimizer .step ()

    return model 

def evaluate_pytorch_model (model ,X_test ,label_encoder ):
    model .eval ()
    with torch .no_grad ():
        inputs =torch .from_numpy (X_test .astype (np .float32 ))
        outputs =model (inputs )
        _ ,predicted =torch .max (outputs .data ,1 )
        y_pred_encoded =predicted .numpy ()
        y_pred =label_encoder .inverse_transform (y_pred_encoded )
    return y_pred 


def train_tensorflow_dnn_model (X ,y ,input_size ,hidden_size ,num_classes ,num_epochs ,batch_size ,learning_rate ):
    model =tf .keras .Sequential ([
    tf .keras .layers .Dense (hidden_size ,activation ='relu',input_shape =(input_size ,)),
    tf .keras .layers .Dense (hidden_size ,activation ='relu'),
    tf .keras .layers .Dense (num_classes ,activation ='softmax')
    ])

    model .compile (optimizer =tf .keras .optimizers .Adam (learning_rate =learning_rate ),
    loss ='sparse_categorical_crossentropy',
    metrics =['accuracy'])

    return model 


def train_tensorflow_cnn_model (X ,y ,input_size ,num_classes ,num_epochs ,batch_size ,learning_rate ):
    model =tf .keras .Sequential ([
    tf .keras .layers .Conv1D (filters =32 ,kernel_size =3 ,activation ='relu',input_shape =(input_size ,1 )),
    tf .keras .layers .MaxPooling1D (pool_size =2 ),
    tf .keras .layers .Conv1D (filters =64 ,kernel_size =3 ,activation ='relu'),
    tf .keras .layers .MaxPooling1D (pool_size =2 ),
    tf .keras .layers .Flatten (),
    tf .keras .layers .Dense (128 ,activation ='relu'),
    tf .keras .layers .Dense (num_classes ,activation ='softmax')
    ])

    model .compile (optimizer =tf .keras .optimizers .Adam (learning_rate =learning_rate ),
    loss ='sparse_categorical_crossentropy',
    metrics =['accuracy'])

    return model 


def train_tensorflow_rnn_model (X ,y ,input_size ,hidden_size ,num_layers ,num_classes ,num_epochs ,batch_size ,learning_rate ):
    model =tf .keras .Sequential ([
    tf .keras .layers .LSTM (hidden_size ,return_sequences =True ,input_shape =(1 ,input_size )),
    *[tf .keras .layers .LSTM (hidden_size ,return_sequences =True )for _ in range (num_layers -2 )],
    tf .keras .layers .LSTM (hidden_size ),
    tf .keras .layers .Dense (num_classes ,activation ='softmax')
    ])

    model .compile (optimizer =tf .keras .optimizers .Adam (learning_rate =learning_rate ),
    loss ='sparse_categorical_crossentropy',
    metrics =['accuracy'])

    return model 


def train_tensorflow_capsule_model (X ,y ,input_size ,num_classes ,num_epochs ,batch_size ,learning_rate ):
    input_layer =tf .keras .layers .Input (shape =(input_size ,))
    reshape_layer =tf .keras .layers .Reshape ((input_size ,1 ,1 ))(input_layer )
    conv1 =tf .keras .layers .Conv2D (filters =256 ,kernel_size =(input_size ,1 ),strides =1 ,activation ='relu')(reshape_layer )
    primary_caps =tf .keras .layers .Conv2D (filters =32 ,kernel_size =1 ,strides =1 ,activation ='relu')(conv1 )
    primary_caps =tf .keras .layers .Reshape (target_shape =(-1 ,8 ))(primary_caps )
    primary_caps =tf .keras .layers .Lambda (squash )(primary_caps )
    digit_caps =capsule_layer (num_capsules =num_classes ,dim_capsule =16 ,routings =3 )(primary_caps )
    output =tf .keras .layers .Lambda (lambda x :tf .sqrt (tf .reduce_sum (tf .square (x ),axis =-1 ,keepdims =True )))(digit_caps )
    output =tf .keras .layers .Reshape (target_shape =(-1 ,))(output )
    model =tf .keras .models .Model (inputs =input_layer ,outputs =output )
    model .compile (optimizer =tf .keras .optimizers .Adam (learning_rate =learning_rate ),
    loss ='sparse_categorical_crossentropy',
    metrics =['accuracy'])
    return model 


def squash (vectors ,axis =-1 ):
    s_squared_norm =tf .reduce_sum (tf .square (vectors ),axis =axis ,keepdims =True )
    scale =s_squared_norm /(1 +s_squared_norm )/tf .sqrt (s_squared_norm +1e-8 )
    return scale *vectors 


def capsule_layer (num_capsules ,dim_capsule ,routings =3 ):
    def capsule (inputs ):
        input_shape =tf .shape (inputs )
        inputs =tf .reshape (inputs ,(input_shape [0 ],input_shape [1 ],1 ,input_shape [2 ]))
        inputs =tf .tile (inputs ,[1 ,1 ,num_capsules ,1 ])
        inputs =tf .reshape (inputs ,(-1 ,input_shape [1 ],num_capsules ,dim_capsule ))

        b =tf .zeros_like (inputs [:,:,:,0 ])

        for i in range (routings ):
            c =tf .nn .softmax (b ,axis =1 )
            outputs =squash (tf .matmul (c ,inputs ))

            if i <routings -1 :
                b +=tf .matmul (outputs ,inputs ,transpose_b =True )

        return tf .reshape (outputs ,(-1 ,num_capsules ,dim_capsule ))

    return tf .keras .layers .Lambda (capsule )


def train_and_evaluate_classifiers (X ,y ,selected_classifiers ,k ,n_estimators ,random_state ,hidden_layer_sizes ,criterion ,n_jobs ,shuffle ,num_layers ,num_epochs ,batch_size ,learning_rate ):
    results ={}

    for classifier_name in selected_classifiers :
        if classifier_name =="Deep Neural Networks":

            label_encoder =LabelEncoder ()
            y =label_encoder .fit_transform (y )

            input_size =X .shape [1 ]
            num_classes =len (np .unique (y ))

            model =train_tensorflow_dnn_model (X ,y ,input_size ,hidden_layer_sizes [0 ],num_classes ,num_epochs ,batch_size ,learning_rate )


            kf =KFold (n_splits =k ,shuffle =shuffle ,random_state =random_state )
            accuracies =[]
            f1_scores =[]
            sensitivities =[]
            specificities =[]
            confusion_matrices =[]





            for train_index ,test_index in kf .split (X ):
                X_train ,X_test =X [train_index ],X [test_index ]
                y_train ,y_test =y [train_index ],y [test_index ]

                model .fit (X_train ,y_train ,epochs =num_epochs ,batch_size =batch_size ,verbose =0 )
                y_pred =np .argmax (model .predict (X_test ),axis =1 )

                accuracy =accuracy_score (y_test ,y_pred )
                accuracies .append (accuracy )

                f1 =f1_score (y_test ,y_pred ,average ='weighted')
                f1_scores .append (f1 )

                sensitivity =recall_score (y_test ,y_pred ,average ='weighted')
                sensitivities .append (sensitivity )

                specificity =precision_score (y_test ,y_pred ,average ='weighted')
                specificities .append (specificity )

                cm =confusion_matrix (y_test ,y_pred )
                confusion_matrices .append (cm )

            avg_accuracy =np .mean (accuracies )
            avg_f1_score =np .mean (f1_scores )
            avg_sensitivity =np .mean (sensitivities )
            avg_specificity =np .mean (specificities )
            avg_confusion_matrix =np .mean (confusion_matrices ,axis =0 )

            results [classifier_name ]={
            'model':model ,
            'avg_accuracy':avg_accuracy ,
            'avg_f1_score':avg_f1_score ,
            'avg_sensitivity':avg_sensitivity ,
            'avg_specificity':avg_specificity ,
            'avg_confusion_matrix':avg_confusion_matrix 
            }

        elif classifier_name =="Convolutional Neural Networks":
            label_encoder =LabelEncoder ()
            y =label_encoder .fit_transform (y )

            input_size =X .shape [1 ]
            num_classes =len (np .unique (y ))

            model =train_tensorflow_cnn_model (X ,y ,input_size ,num_classes ,num_epochs ,batch_size ,learning_rate )


            kf =KFold (n_splits =k ,shuffle =shuffle ,random_state =random_state )
            accuracies =[]
            f1_scores =[]
            sensitivities =[]
            specificities =[]
            confusion_matrices =[]

            for train_index ,test_index in kf .split (X ):
                X_train ,X_test =X [train_index ],X [test_index ]
                y_train ,y_test =y [train_index ],y [test_index ]

                model .fit (X_train .reshape (-1 ,input_size ,1 ),y_train ,epochs =num_epochs ,batch_size =batch_size ,verbose =0 )
                y_pred =np .argmax (model .predict (X_test .reshape (-1 ,input_size ,1 )),axis =1 )

                accuracy =accuracy_score (y_test ,y_pred )
                accuracies .append (accuracy )

                f1 =f1_score (y_test ,y_pred ,average ='weighted')
                f1_scores .append (f1 )

                sensitivity =recall_score (y_test ,y_pred ,average ='weighted')
                sensitivities .append (sensitivity )

                specificity =precision_score (y_test ,y_pred ,average ='weighted')
                specificities .append (specificity )

                cm =confusion_matrix (y_test ,y_pred )
                confusion_matrices .append (cm )

            avg_accuracy =np .mean (accuracies )
            avg_f1_score =np .mean (f1_scores )
            avg_sensitivity =np .mean (sensitivities )
            avg_specificity =np .mean (specificities )
            avg_confusion_matrix =np .mean (confusion_matrices ,axis =0 )

            results [classifier_name ]={
            'model':model ,
            'avg_accuracy':avg_accuracy ,
            'avg_f1_score':avg_f1_score ,
            'avg_sensitivity':avg_sensitivity ,
            'avg_specificity':avg_specificity ,
            'avg_confusion_matrix':avg_confusion_matrix 
            }

        elif classifier_name =="Recurrent Neural Networks":
            label_encoder =LabelEncoder ()
            y =label_encoder .fit_transform (y )

            input_size =X .shape [1 ]
            hidden_size =hidden_layer_sizes [0 ]
            num_classes =len (np .unique (y ))

            model =train_tensorflow_rnn_model (X ,y ,input_size ,hidden_size ,num_layers ,num_classes ,num_epochs ,batch_size ,learning_rate )


            kf =KFold (n_splits =k ,shuffle =shuffle ,random_state =random_state )
            accuracies =[]
            f1_scores =[]
            sensitivities =[]
            specificities =[]
            confusion_matrices =[]

            for train_index ,test_index in kf .split (X ):
                X_train ,X_test =X [train_index ],X [test_index ]
                y_train ,y_test =y [train_index ],y [test_index ]

                model .fit (X_train .reshape (-1 ,1 ,input_size ),y_train ,epochs =num_epochs ,batch_size =batch_size ,verbose =0 )
                y_pred =np .argmax (model .predict (X_test .reshape (-1 ,1 ,input_size )),axis =1 )

                accuracy =accuracy_score (y_test ,y_pred )
                accuracies .append (accuracy )

                f1 =f1_score (y_test ,y_pred ,average ='weighted')
                f1_scores .append (f1 )

                sensitivity =recall_score (y_test ,y_pred ,average ='weighted')
                sensitivities .append (sensitivity )

                specificity =precision_score (y_test ,y_pred ,average ='weighted')
                specificities .append (specificity )

                cm =confusion_matrix (y_test ,y_pred )
                confusion_matrices .append (cm )

            avg_accuracy =np .mean (accuracies )
            avg_f1_score =np .mean (f1_scores )
            avg_sensitivity =np .mean (sensitivities )
            avg_specificity =np .mean (specificities )
            avg_confusion_matrix =np .mean (confusion_matrices ,axis =0 )

            results [classifier_name ]={
            'model':model ,
            'avg_accuracy':avg_accuracy ,
            'avg_f1_score':avg_f1_score ,
            'avg_sensitivity':avg_sensitivity ,
            'avg_specificity':avg_specificity ,
            'avg_confusion_matrix':avg_confusion_matrix 
            }

        elif classifier_name =="Capsule Networks":
            label_encoder =LabelEncoder ()
            y =label_encoder .fit_transform (y )

            input_size =X .shape [1 ]
            num_classes =len (np .unique (y ))

            model =train_tensorflow_capsule_model (X ,y ,input_size ,num_classes ,num_epochs ,batch_size ,learning_rate )


            kf =KFold (n_splits =k ,shuffle =shuffle ,random_state =random_state )
            accuracies =[]
            f1_scores =[]
            sensitivities =[]
            specificities =[]
            confusion_matrices =[]

            for train_index ,test_index in kf .split (X ):
                X_train ,X_test =X [train_index ],X [test_index ]
                y_train ,y_test =y [train_index ],y [test_index ]

                model .fit (X_train ,y_train ,epochs =num_epochs ,batch_size =batch_size ,verbose =0 )
                y_pred =np .argmax (model .predict (X_test ),axis =1 )

                accuracy =accuracy_score (y_test ,y_pred )
                accuracies .append (accuracy )

                f1 =f1_score (y_test ,y_pred ,average ='weighted')
                f1_scores .append (f1 )

                sensitivity =recall_score (y_test ,y_pred ,average ='weighted')
                sensitivities .append (sensitivity )

                specificity =precision_score (y_test ,y_pred ,average ='weighted')
                specificities .append (specificity )

                cm =confusion_matrix (y_test ,y_pred )
                confusion_matrices .append (cm )

            avg_accuracy =np .mean (accuracies )
            avg_f1_score =np .mean (f1_scores )
            avg_sensitivity =np .mean (sensitivities )
            avg_specificity =np .mean (specificities )
            avg_confusion_matrix =np .mean (confusion_matrices ,axis =0 )

            results [classifier_name ]={
            'model':model ,
            'avg_accuracy':avg_accuracy ,
            'avg_f1_score':avg_f1_score ,
            'avg_sensitivity':avg_sensitivity ,
            'avg_specificity':avg_specificity ,
            'avg_confusion_matrix':avg_confusion_matrix 
            }

        elif classifier_name .startswith ("RAPIDS"):

            if classifier_name =="RAPIDS RandomForestClassifier":
                classifier =cuml .ensemble .RandomForestClassifier (n_estimators =n_estimators ,random_state =random_state ,n_jobs =n_jobs )
            elif classifier_name =="RAPIDS GradientBoostingClassifier":
                classifier =cuml .ensemble .GradientBoostingClassifier (n_estimators =n_estimators ,random_state =random_state )
            elif classifier_name =="RAPIDS XGBoostClassifier":
                classifier =cuml .XGBClassifier (n_estimators =n_estimators ,random_state =random_state ,n_jobs =n_jobs )


            kf =KFold (n_splits =k ,shuffle =shuffle ,random_state =random_state )
            accuracies =[]
            f1_scores =[]
            sensitivities =[]
            specificities =[]
            confusion_matrices =[]
            feature_importances_list =[]

            for train_index ,test_index in kf .split (X ):
                X_train ,X_test =X [train_index ],X [test_index ]
                y_train ,y_test =y [train_index ],y [test_index ]

                classifier .fit (X_train ,y_train )
                y_pred =classifier .predict (X_test )

                accuracy =accuracy_score (y_test ,y_pred )
                accuracies .append (accuracy )

                f1 =f1_score (y_test ,y_pred ,average ='weighted')
                f1_scores .append (f1 )

                sensitivity =recall_score (y_test ,y_pred ,average ='weighted')
                sensitivities .append (sensitivity )

                specificity =precision_score (y_test ,y_pred ,average ='weighted')
                specificities .append (specificity )

                cm =confusion_matrix (y_test ,y_pred )
                confusion_matrices .append (cm )

                if hasattr (classifier ,'feature_importances_'):
                    feature_importances_list .append (classifier .feature_importances_ )

            avg_accuracy =np .mean (accuracies )
            avg_f1_score =np .mean (f1_scores )
            avg_sensitivity =np .mean (sensitivities )
            avg_specificity =np .mean (specificities )
            avg_confusion_matrix =np .mean (confusion_matrices ,axis =0 )

            if feature_importances_list :
                avg_feature_importances =np .mean (feature_importances_list ,axis =0 )
            else :
                avg_feature_importances =None 

            results [classifier_name ]={
            'model':classifier ,
            'avg_accuracy':avg_accuracy ,
            'avg_f1_score':avg_f1_score ,
            'avg_sensitivity':avg_sensitivity ,
            'avg_specificity':avg_specificity ,
            'avg_confusion_matrix':avg_confusion_matrix ,
            'avg_feature_importances':avg_feature_importances 
            }

        else :

            classifier ={
            "RandomForestClassifier":RandomForestClassifier (n_estimators =n_estimators ,random_state =random_state ,n_jobs =n_jobs ,criterion =criterion ),
            "GradientBoostingClassifier":GradientBoostingClassifier (n_estimators =n_estimators ,random_state =random_state ),
            "LGBMClassifier":LGBMClassifier (n_estimators =n_estimators ,random_state =random_state ,n_jobs =n_jobs ),
            "AdaBoostClassifier":AdaBoostClassifier (n_estimators =n_estimators ,random_state =random_state ),
            "DecisionTreeClassifier":DecisionTreeClassifier (random_state =random_state ,criterion =criterion ),
            "ExtraTreesClassifier":ExtraTreesClassifier (n_estimators =n_estimators ,random_state =random_state ,n_jobs =n_jobs ,criterion =criterion ),
            "SVC":SVC (kernel ='rbf',random_state =random_state ),
            "KNeighborsClassifier":KNeighborsClassifier (n_neighbors =5 ),
            "LogisticRegression":LogisticRegression (random_state =random_state ),
            "GaussianNB":GaussianNB (),
            "MLPClassifier":MLPClassifier (hidden_layer_sizes =(100 ,),random_state =random_state )
            }[classifier_name ]


            kf =KFold (n_splits =k ,shuffle =shuffle ,random_state =random_state )
            accuracies =[]
            f1_scores =[]
            sensitivities =[]
            specificities =[]
            confusion_matrices =[]
            feature_importances_list =[]
            y_true_all =[]
            y_pred_all =[]

            for train_index ,test_index in kf .split (X ):
                X_train ,X_test =X [train_index ],X [test_index ]
                y_train ,y_test =y [train_index ],y [test_index ]

                classifier .fit (X_train ,y_train )
                y_pred =classifier .predict (X_test )

                accuracy =accuracy_score (y_test ,y_pred )
                accuracies .append (accuracy )

                f1 =f1_score (y_test ,y_pred ,average ='weighted')
                f1_scores .append (f1 )

                sensitivity =recall_score (y_test ,y_pred ,average ='weighted')
                sensitivities .append (sensitivity )

                specificity =precision_score (y_test ,y_pred ,average ='weighted')
                specificities .append (specificity )

                y_true_all .extend (y_test )
                y_pred_all .extend (y_pred )




                if hasattr (classifier ,'feature_importances_'):
                    feature_importances_list .append (classifier .feature_importances_ )
                elif hasattr (classifier ,'coef_'):
                    feature_importances_list .append (classifier .coef_ .ravel ())

            avg_accuracy =np .mean (accuracies )
            avg_f1_score =np .mean (f1_scores )
            avg_sensitivity =np .mean (sensitivities )
            avg_specificity =np .mean (specificities )

            avg_confusion_matrix =confusion_matrix (y_true_all ,y_pred_all )

            if feature_importances_list :
                avg_feature_importances =np .abs (np .mean (feature_importances_list ,axis =0 ))
            else :
                avg_feature_importances =None 

            results [classifier_name ]={
            'model':classifier ,
            'avg_accuracy':avg_accuracy ,
            'avg_f1_score':avg_f1_score ,
            'avg_sensitivity':avg_sensitivity ,
            'avg_specificity':avg_specificity ,
            'avg_confusion_matrix':avg_confusion_matrix ,
            'avg_feature_importances':avg_feature_importances 
            }

    return results 

class ResultsWidget (QWidget ):
    def __init__ (self ,classifier_name ):
        super ().__init__ ()

        layout =QGridLayout (self )


        self .confusion_matrix_plot =FigureCanvas (Figure (figsize =(5 ,5 )))
        self .confusion_matrix_toolbar =NavigationToolbar (self .confusion_matrix_plot ,self )
        layout .addWidget (self .confusion_matrix_plot ,0 ,0 )
        layout .addWidget (self .confusion_matrix_toolbar ,1 ,0 )

        self .feature_importance_plot =FigureCanvas (Figure (figsize =(5 ,5 )))
        self .feature_importance_toolbar =NavigationToolbar (self .feature_importance_plot ,self )
        layout .addWidget (self .feature_importance_plot ,0 ,1 )
        layout .addWidget (self .feature_importance_toolbar ,1 ,1 )

        self .test_confusion_matrix_plot =FigureCanvas (Figure (figsize =(5 ,5 )))
        self .test_confusion_matrix_toolbar =NavigationToolbar (self .test_confusion_matrix_plot ,self )
        layout .addWidget (self .test_confusion_matrix_plot ,2 ,0 )
        layout .addWidget (self .test_confusion_matrix_toolbar ,3 ,0 )


        self .info_text_edit =QTextEdit ()
        self .info_text_edit .setReadOnly (True )
        self .info_text_edit .setWordWrapMode (QTextOption .WrapMode .WrapAtWordBoundaryOrAnywhere )
        layout .addWidget (self .info_text_edit ,2 ,1 )

    def update_plots (self ,confusion_matrix ,feature_importances ,test_confusion_matrix ,labels ):


        if confusion_matrix is not None :
            self .confusion_matrix_plot .figure .clear ()
            ax =self .confusion_matrix_plot .figure .add_subplot (111 )

            confusion_matrix_norm =confusion_matrix .astype ('float')/confusion_matrix .sum (axis =1 )[:,np .newaxis ]

            im =ax .imshow (confusion_matrix_norm ,cmap ='Blues',interpolation ='nearest')
            ax .set_title ("Average Confusion Matrix")
            ax .set_xlabel ("Predicted Label")
            ax .set_ylabel ("True Label")
            ax .set_xticks (np .arange (len (labels )))
            ax .set_yticks (np .arange (len (labels )))
            ax .set_xticklabels (labels )
            ax .set_yticklabels (labels )


            for i in range (len (labels )):
                for j in range (len (labels )):
                    ax .text (j ,i ,f"{confusion_matrix_norm[i, j]*100:.1f}%",ha ='center',va ='center',color ='black')

            self .confusion_matrix_plot .figure .colorbar (im )
            self .confusion_matrix_plot .draw ()
        elif self .confusion_matrix_plot .figure .get_axes ():
            pass 


        if feature_importances is not None :
            self .feature_importance_plot .figure .clear ()
            ax =self .feature_importance_plot .figure .add_subplot (111 )
            ax .bar (range (len (feature_importances )),feature_importances )
            ax .set_title ("Feature Importances")
            ax .set_xlabel ("Feature")
            ax .set_ylabel ("Importance")
            self .feature_importance_plot .draw ()
        elif self .feature_importance_plot .figure .get_axes ():
            pass 


        self .test_confusion_matrix_plot .figure .clear ()
        ax =self .test_confusion_matrix_plot .figure .add_subplot (111 )

        if test_confusion_matrix is not None :

            test_confusion_matrix_norm =test_confusion_matrix .astype ('float')/test_confusion_matrix .sum (axis =1 )[:,np .newaxis ]
            test_confusion_matrix_norm =np .nan_to_num (test_confusion_matrix_norm ,nan =0.0 )

            im =ax .imshow (test_confusion_matrix_norm ,cmap ='Blues',interpolation ='nearest')
            ax .set_title ("Test Confusion Matrix")
            ax .set_xlabel ("Predicted Label")
            ax .set_ylabel ("True Label")
            ax .set_xticks (np .arange (len (labels )))
            ax .set_yticks (np .arange (len (labels )))
            ax .set_xticklabels (labels )
            ax .set_yticklabels (labels )


            for i in range (len (labels )):
                for j in range (len (labels )):
                    ax .text (j ,i ,f"{test_confusion_matrix_norm[i, j]*100:.1f}%",ha ='center',va ='center',color ='black')

            self .test_confusion_matrix_plot .figure .colorbar (im )
        else :
            ax .text (0.5 ,0.5 ,"Test confusion matrix not available",horizontalalignment ='center',verticalalignment ='center')
        self .test_confusion_matrix_plot .draw ()

    def update_info (self ,info_text ):
        self .info_text_edit .append (info_text )

class MainWindow (QMainWindow ):
    def __init__ (self ):
        super ().__init__ ()
        self .setWindowTitle ("SD ML Training and Testing Tool")


        main_widget =QWidget ()
        main_layout =QHBoxLayout (main_widget )
        self .setCentralWidget (main_widget )


        splitter =QSplitter ()
        main_layout .addWidget (splitter )


        left_widget =QWidget ()
        right_widget =QWidget ()
        splitter .addWidget (left_widget )
        splitter .addWidget (right_widget )
        splitter .setSizes ([300 ,700 ])


        left_layout =QVBoxLayout (left_widget )

        self .app_name_label =QLabel ("SD ML Training and Testing Tool")
        self .app_name_label .setAlignment (Qt .AlignmentFlag .AlignCenter )
        self .app_name_label .setStyleSheet ("font-size: 22px; font-weight: bold;")
        self .app_name_label .setWordWrap (True )
        left_layout .addWidget (self .app_name_label )
        self .email_label =QLabel ("shankar.dutt@anu.edu.au")
        self .email_label .setAlignment (Qt .AlignmentFlag .AlignCenter )
        self .email_label .setWordWrap (True )
        left_layout .addWidget (self .email_label )
        left_tab_widget =QTabWidget ()
        left_layout .addWidget (left_tab_widget )


        general_tab =QWidget ()
        general_tab_layout =QVBoxLayout (general_tab )
        settings_tab =QWidget ()
        settings_tab_layout =QVBoxLayout (settings_tab )
        left_tab_widget .addTab (general_tab ,"General")
        left_tab_widget .addTab (settings_tab ,"Settings")


        general_scroll_area =QScrollArea ()
        general_scroll_area .setWidgetResizable (True )
        general_scroll_widget =QWidget ()
        general_scroll_area .setWidget (general_scroll_widget )
        general_layout =QVBoxLayout (general_scroll_widget )
        general_tab_layout .addWidget (general_scroll_area )


        self .classifier_type_combo =QComboBox ()
        self .classifier_type_combo .addItem ("Train New Classifier")
        self .classifier_type_combo .addItem ("Load Trained Classifier")
        general_layout .addWidget (self .classifier_type_combo )

        self .training_sets_group =QGroupBox ("Training Sets")
        training_sets_layout =QVBoxLayout ()
        self .num_training_sets_combo =QComboBox ()
        for i in range (2 ,11 ):
            self .num_training_sets_combo .addItem (str (i ))
        training_sets_layout .addWidget (self .num_training_sets_combo )
        self .training_file_buttons =[]
        self .training_file_labels =[]
        training_sets_layout .addStretch (1 )
        self .training_sets_group .setLayout (training_sets_layout )
        general_layout .addWidget (self .training_sets_group )

        self .classifier_group =QGroupBox ("Classifier Options")
        classifier_layout =QVBoxLayout ()
        self .classifier_checkboxes =[]
        classifier_layout .addStretch (1 )
        self .classifier_group .setLayout (classifier_layout )
        general_layout .addWidget (self .classifier_group )

        self .save_classifier_checkbox =QCheckBox ("Save Trained Classifier")
        general_layout .addWidget (self .save_classifier_checkbox )

        self .train_button =QPushButton ("Train Classifier")
        general_layout .addWidget (self .train_button )

        self .test_group =QGroupBox ("Testing")
        test_layout =QVBoxLayout ()
        self .select_test_button =QPushButton ("Select Test/Unknown Dataset")
        test_layout .addWidget (self .select_test_button )

        self .test_file_label =QTextEdit ()
        self .test_file_label .setReadOnly (True )
        self .test_file_label .setWordWrapMode (QTextOption .WrapMode .WrapAtWordBoundaryOrAnywhere )
        self .test_file_label .setFixedHeight (60 )
        test_layout .addWidget (self .test_file_label )

        self .perform_test_button =QPushButton ("Perform Test")
        test_layout .addWidget (self .perform_test_button )
        test_layout .addStretch (1 )
        self .test_group .setLayout (test_layout )
        general_layout .addWidget (self .test_group )


        general_settings_group =QGroupBox ("General Settings")
        cpu_settings_group =QGroupBox ("CPU Settings")
        gpu_settings_group =QGroupBox ("GPU Settings")


        general_settings_layout =QVBoxLayout ()
        cpu_settings_layout =QVBoxLayout ()
        gpu_settings_layout =QVBoxLayout ()


        general_settings_layout .addWidget (QLabel ("Number of K-Fold Cross Validation:"))
        self .k_fold_combo =QComboBox ()
        for i in range (2 ,11 ):
            self .k_fold_combo .addItem (str (i ))
        self .k_fold_combo .setCurrentText ("5")
        general_settings_layout .addWidget (self .k_fold_combo )

        general_settings_layout .addWidget (QLabel ("Random State:"))
        self .random_state_spinbox =QSpinBox ()
        self .random_state_spinbox .setMinimum (0 )
        self .random_state_spinbox .setMaximum (100000 )
        self .random_state_spinbox .setValue (42 )
        general_settings_layout .addWidget (self .random_state_spinbox )

        general_settings_layout .addWidget (QLabel ("Shuffle:"))
        self .shuffle_combo =QComboBox ()
        self .shuffle_combo .addItem ("True")
        self .shuffle_combo .addItem ("False")
        self .shuffle_combo .setCurrentText ("True")
        general_settings_layout .addWidget (self .shuffle_combo )

        general_settings_group .setLayout (general_settings_layout )


        cpu_settings_layout .addWidget (QLabel ("Number of Estimators:"))
        self .num_estimators_combo =QComboBox ()
        self .num_estimators_combo .addItem ("100")
        self .num_estimators_combo .addItem ("500")
        self .num_estimators_combo .addItem ("1000")
        self .num_estimators_combo .setCurrentText ("1000")
        cpu_settings_layout .addWidget (self .num_estimators_combo )

        cpu_settings_layout .addWidget (QLabel ("Criterion:"))
        self .criterion_combo =QComboBox ()
        self .criterion_combo .addItem ("gini")
        self .criterion_combo .addItem ("entropy")
        self .criterion_combo .setCurrentText ("gini")
        cpu_settings_layout .addWidget (self .criterion_combo )

        cpu_settings_layout .addWidget (QLabel ("Number of Jobs:"))
        self .n_jobs_combo =QComboBox ()
        self .n_jobs_combo .addItem ("1")
        self .n_jobs_combo .addItem ("2")
        self .n_jobs_combo .addItem ("4")
        self .n_jobs_combo .addItem ("-1")
        self .n_jobs_combo .setCurrentText ("-1")
        cpu_settings_layout .addWidget (self .n_jobs_combo )

        cpu_settings_group .setLayout (cpu_settings_layout )


        gpu_settings_layout .addWidget (QLabel ("Hidden Layer Sizes:"))
        self .hidden_layer_sizes_spinbox =QSpinBox ()
        self .hidden_layer_sizes_spinbox .setMinimum (1 )
        self .hidden_layer_sizes_spinbox .setMaximum (10000 )
        self .hidden_layer_sizes_spinbox .setValue (1000 )
        gpu_settings_layout .addWidget (self .hidden_layer_sizes_spinbox )

        gpu_settings_layout .addWidget (QLabel ("Number of Layers:"))
        self .num_layers_spinbox =QSpinBox ()
        self .num_layers_spinbox .setMinimum (1 )
        self .num_layers_spinbox .setMaximum (10 )
        self .num_layers_spinbox .setValue (2 )
        gpu_settings_layout .addWidget (self .num_layers_spinbox )

        gpu_settings_layout .addWidget (QLabel ("Number of Epochs:"))
        self .num_epochs_spinbox =QSpinBox ()
        self .num_epochs_spinbox .setMinimum (1 )
        self .num_epochs_spinbox .setMaximum (1000 )
        self .num_epochs_spinbox .setValue (50 )
        gpu_settings_layout .addWidget (self .num_epochs_spinbox )

        gpu_settings_layout .addWidget (QLabel ("Batch Size:"))
        self .batch_size_spinbox =QSpinBox ()
        self .batch_size_spinbox .setMinimum (1 )
        self .batch_size_spinbox .setMaximum (1000 )
        self .batch_size_spinbox .setValue (32 )
        gpu_settings_layout .addWidget (self .batch_size_spinbox )

        gpu_settings_layout .addWidget (QLabel ("Learning Rate:"))
        self .learning_rate_combo =QComboBox ()
        self .learning_rate_combo .addItem ("0.001")
        self .learning_rate_combo .addItem ("0.01")
        self .learning_rate_combo .addItem ("0.1")
        self .learning_rate_combo .setCurrentText ("0.001")
        gpu_settings_layout .addWidget (self .learning_rate_combo )

        gpu_settings_group .setLayout (gpu_settings_layout )


        settings_tab_layout .addWidget (general_settings_group )
        settings_tab_layout .addWidget (cpu_settings_group )
        settings_tab_layout .addWidget (gpu_settings_group )
        settings_tab_layout .addStretch (1 )


        right_layout =QVBoxLayout (right_widget )
        self .right_tab_widget =QTabWidget ()
        scroll_area =QScrollArea ()
        scroll_area .setWidgetResizable (True )
        scroll_area .setWidget (self .right_tab_widget )
        right_layout .addWidget (scroll_area )


        self .classifier_type_combo .currentIndexChanged .connect (self .update_classifier_options )
        self .num_training_sets_combo .currentIndexChanged .connect (self .update_training_file_buttons )
        self .train_button .clicked .connect (self .train_classifier )
        self .select_test_button .clicked .connect (self .select_test_dataset )
        self .perform_test_button .clicked .connect (self .perform_test )


        self .update_classifier_options ()
        self .update_training_file_buttons ()


        self .trained_classifiers ={}
        self .loaded_classifiers ={}

    def update_classifier_options (self ):
        classifier_type =self .classifier_type_combo .currentText ()


        for checkbox in self .classifier_checkboxes :
            checkbox .deleteLater ()
        self .classifier_checkboxes .clear ()


        if classifier_type =="Train New Classifier":
            self .training_sets_group .setTitle ("Training Sets")
            self .save_classifier_checkbox .setVisible (True )
            self .train_button .setVisible (True )
            self .classifier_group .setVisible (True )


            self .num_training_sets_combo .clear ()
            for i in range (2 ,11 ):
                self .num_training_sets_combo .addItem (str (i ))
            self .num_training_sets_combo .setCurrentIndex (0 )

        else :
            self .training_sets_group .setTitle ("Trained Classifiers")
            self .save_classifier_checkbox .setVisible (False )
            self .train_button .setVisible (False )
            self .classifier_group .setVisible (False )


            self .num_training_sets_combo .clear ()
            for i in range (1 ,11 ):
                self .num_training_sets_combo .addItem (str (i ))
            self .num_training_sets_combo .setCurrentIndex (0 )


        if classifier_type =="Train New Classifier":

            cpu_classifiers =[
            "RandomForestClassifier",
            "GradientBoostingClassifier",
            "LGBMClassifier",
            "AdaBoostClassifier",
            "DecisionTreeClassifier",
            "ExtraTreesClassifier",
            "SVC",
            "KNeighborsClassifier",
            "LogisticRegression",
            "GaussianNB",
            "MLPClassifier"
            ]


            tensor_classifiers =[
            "Deep Neural Networks",
            "Convolutional Neural Networks",
            "Recurrent Neural Networks",
            "Capsule Networks"
            ]


            gpu_classifiers =[]
            if torch .cuda .is_available ():
                gpu_classifiers =[
                "RAPIDS RandomForestClassifier",
                "RAPIDS GradientBoostingClassifier",
                "RAPIDS XGBoostClassifier"
                ]


            cpu_group =QGroupBox ("CPU-based Classifiers")
            cpu_layout =QVBoxLayout ()
            for classifier in cpu_classifiers :
                checkbox =QCheckBox (classifier )
                self .classifier_checkboxes .append (checkbox )
                cpu_layout .addWidget (checkbox )
            cpu_group .setLayout (cpu_layout )

            tensor_group =QGroupBox ("Tensorflow-based Classifiers (GPU/CPU based on the system)")
            tensor_layout =QVBoxLayout ()
            for classifier in tensor_classifiers :
                checkbox =QCheckBox (classifier )
                self .classifier_checkboxes .append (checkbox )
                tensor_layout .addWidget (checkbox )
            tensor_group .setLayout (tensor_layout )

            gpu_group =QGroupBox ("Nvidia GPU-based Classifiers (RAPIDS)")
            gpu_layout =QVBoxLayout ()
            for classifier in gpu_classifiers :
                checkbox =QCheckBox (classifier )
                self .classifier_checkboxes .append (checkbox )
                gpu_layout .addWidget (checkbox )
            gpu_group .setLayout (gpu_layout )


            self .classifier_group .layout ().addWidget (cpu_group )
            self .classifier_group .layout ().addWidget (tensor_group )
            self .classifier_group .layout ().addWidget (gpu_group )
            self .classifier_group .layout ().addStretch (1 )

        self .update_training_file_buttons ()

    def update_training_file_buttons (self ):
        current_text =self .num_training_sets_combo .currentText ()
        if current_text :
            num_sets =int (current_text )
        else :
            num_sets =0 


        for button in self .training_file_buttons :
            button .deleteLater ()
        for label in self .training_file_labels :
            label .deleteLater ()

        self .training_file_buttons .clear ()
        self .training_file_labels .clear ()


        for i in range (num_sets ):
            if self .classifier_type_combo .currentText ()=="Train New Classifier":
                button =QPushButton (f"Select Training File {i+1}")
                button .clicked .connect (lambda _ ,idx =i :self .select_training_file (idx ))
            else :
                button =QPushButton (f"Select Trained Classifier {i+1}")
                button .setProperty ("index",i )
                button .clicked .connect (self .select_trained_classifier )

            self .training_file_buttons .append (button )

            label =QTextEdit ()
            self .training_file_labels .append (label )


        for button ,label in zip (self .training_file_buttons ,self .training_file_labels ):
            self .training_sets_group .layout ().insertWidget (self .training_sets_group .layout ().count ()-1 ,button )
            self .training_sets_group .layout ().insertWidget (self .training_sets_group .layout ().count ()-1 ,label )

    def select_trained_classifier (self ):
        button =self .sender ()
        index =button .property ("index")
        file_dialog =QFileDialog ()
        file_path ,_ =file_dialog .getOpenFileName (self ,"Select Trained Classifier","","Classifier Files (*.pkl *.h5)")
        if file_path :
            classifier_name =os .path .splitext (os .path .basename (file_path ))[0 ]
            file_extension =os .path .splitext (file_path )[1 ]

            try :
                if file_extension ==".h5":

                    classifier =tf .keras .models .load_model (file_path )
                else :

                    with open (file_path ,'rb')as file :
                        classifier =pickle .load (file )


                if not (hasattr (classifier ,'predict')or hasattr (classifier ,'predict_proba')):
                    raise ValueError ("Invalid classifier object")

                self .loaded_classifiers [classifier_name ]=classifier 
                self .update_loaded_classifiers_info (file_path ,index )
            except (IOError ,pickle .UnpicklingError ,ValueError )as e :
                QMessageBox .critical (self ,"Error",f"Failed to load classifier: {str(e)}")

    def update_loaded_classifiers_info (self ,file_path ,index ):
        self .training_file_labels [index ].clear ()
        self .training_file_labels [index ].append (f"Classifier: {os.path.splitext(os.path.basename(file_path))[0]}")
        self .training_file_labels [index ].append (f"File Path: {file_path}")
        self .training_file_labels [index ].setReadOnly (True )
        self .training_file_labels [index ].setWordWrapMode (QTextOption .WrapMode .WrapAtWordBoundaryOrAnywhere )


    def perform_test (self ):
        if self .classifier_type_combo .currentText ()=="Load Trained Classifier":
            if not hasattr (self ,'loaded_classifiers'):
                QMessageBox .warning (self ,"Warning","No classifiers were loaded.")
                return 
            classifiers =self .loaded_classifiers 
        else :
            classifiers =self .trained_classifiers 
            if not classifiers :
                QMessageBox .warning (self ,"Warning","No classifiers were trained.")
                return 

        if hasattr (self ,'test_file_path'):
            X_test ,y_test_true =load_data (self .test_file_path )

            X_test =np .nan_to_num (X_test ,nan =0.0 )

            for classifier_name ,classifier in classifiers .items ():
                if isinstance (classifier ,(tf .keras .Sequential ,tf .keras .Model )):

                    if classifier_name =="Deep Neural Networks":
                        y_pred =np .argmax (classifier .predict (X_test ),axis =1 )
                    elif classifier_name =="Convolutional Neural Networks":
                        y_pred =np .argmax (classifier .predict (X_test .reshape (-1 ,X_test .shape [1 ],1 )),axis =1 )
                    elif classifier_name =="Recurrent Neural Networks":
                        y_pred =np .argmax (classifier .predict (X_test .reshape (-1 ,1 ,X_test .shape [1 ])),axis =1 )
                    elif classifier_name =="Capsule Networks":
                        y_pred =np .argmax (classifier .predict (X_test .reshape (-1 ,X_test .shape [1 ],1 ,1 )),axis =1 )
                elif classifier_name .startswith ("RAPIDS"):

                    X_test_cudf =cudf .DataFrame (X_test )
                    y_pred =classifier .predict (X_test_cudf )
                    y_pred =y_pred .to_array ()
                else :

                    y_pred =classifier .predict (X_test )

                unique_types ,counts =np .unique (y_pred ,return_counts =True )

                y_mixture =[]
                for biomolecule_type ,count in zip (unique_types ,counts ):
                    y_mixture .extend ([biomolecule_type ]*count )
                y_mixture =np .array (y_mixture )


                test_confusion_matrix =confusion_matrix (y_mixture ,y_pred )

                info_text =f"Biomolecule Type Identification Results for Test Dataset using {classifier_name}:\n"
                for biomolecule_type ,count in zip (unique_types ,counts ):
                    info_text +=f"Biomolecule Type: {biomolecule_type}, Count: {count}\n"

                self .display_results (classifier_name ,None ,None ,test_confusion_matrix ,info_text ,unique_types )


    def train_classifier (self ):

        training_files =[]
        for i ,label in enumerate (self .training_file_labels ):
            if label .toPlainText ():
                training_files .append (label .toPlainText ().split ('\n')[0 ].split (': ')[1 ])


        selected_classifiers =[]
        for checkbox in self .classifier_checkboxes :
            if checkbox .isChecked ():
                selected_classifiers .append (checkbox .text ())


        X ,y =prepare_data (training_files )


        k =int (self .k_fold_combo .currentText ())
        n_estimators =int (self .num_estimators_combo .currentText ())
        random_state =int (self .random_state_spinbox .value ())
        hidden_layer_sizes =(int (self .hidden_layer_sizes_spinbox .value ()),)
        criterion =self .criterion_combo .currentText ()
        n_jobs =int (self .n_jobs_combo .currentText ())
        shuffle =self .shuffle_combo .currentText ()=="True"
        num_layers =int (self .num_layers_spinbox .value ())
        num_epochs =int (self .num_epochs_spinbox .value ())
        batch_size =int (self .batch_size_spinbox .value ())
        learning_rate =float (self .learning_rate_combo .currentText ())


        results =train_and_evaluate_classifiers (X ,y ,selected_classifiers ,k ,n_estimators ,random_state ,hidden_layer_sizes ,criterion ,n_jobs ,shuffle ,num_layers ,num_epochs ,batch_size ,learning_rate )


        self .trained_classifiers ={name :result ['model']for name ,result in results .items ()}


        if self .save_classifier_checkbox .isChecked ():
            file_dialog =QFileDialog ()
            save_path =file_dialog .getExistingDirectory (self ,"Select Save Directory")
            if save_path :
                save_classifiers (results ,save_path )


        for classifier_name ,result in results .items ():
            confusion_matrix =result ['avg_confusion_matrix']
            feature_importances =result .get ('avg_feature_importances')
            labels =np .unique (y )

            info_text =f"Date and Time: {datetime.now()}\n\n"
            info_text +=f"Classifier: {classifier_name}\n\n"
            info_text +=f"Average Accuracy: {result['avg_accuracy']:.4f}\n"
            info_text +=f"Average F1 Score: {result['avg_f1_score']:.4f}\n"
            info_text +=f"Average Sensitivity: {result['avg_sensitivity']:.4f}\n"
            info_text +=f"Average Specificity: {result['avg_specificity']:.4f}\n\n"
            info_text +="Settings:\n"
            info_text +=f"Number of K-Fold Cross Validation: {k}\n"
            info_text +=f"Number of Estimators: {n_estimators}\n"
            info_text +=f"Random State: {random_state}\n"
            info_text +=f"Hidden Layer Sizes: {hidden_layer_sizes}\n"
            info_text +=f"Criterion: {criterion}\n"
            info_text +=f"Number of Jobs: {n_jobs}\n"
            info_text +=f"Shuffle: {shuffle}\n"
            info_text +=f"Number of Layers: {num_layers}\n"
            info_text +=f"Number of Epochs: {num_epochs}\n"
            info_text +=f"Batch Size: {batch_size}\n"
            info_text +=f"Learning Rate: {learning_rate}\n"
            info_text +=f"_____________________________________________________________________ \n\n"

            self .display_results (classifier_name ,confusion_matrix ,feature_importances ,None ,info_text ,labels )

        return results 


    def select_training_file (self ,index ):
        file_dialog =QFileDialog ()
        file_path ,_ =file_dialog .getOpenFileName (self ,"Select Training File","","NPZ Files (*.MLdataset.npz)")

        if file_path :
            X ,label =load_data (file_path )
            self .training_file_labels [index ].setPlainText (f"File: {file_path}\nLabel: {label}")
            self .training_file_labels [index ].setReadOnly (True )
            self .training_file_labels [index ].setWordWrapMode (QTextOption .WrapMode .WrapAtWordBoundaryOrAnywhere )
            self .training_file_labels [index ].setFixedHeight (60 )


    def select_test_dataset (self ):
        file_dialog =QFileDialog ()
        file_path ,_ =file_dialog .getOpenFileName (self ,"Select Test/Unknown Dataset","","NPZ Files (*.MLdataset.npz)")

        if file_path :
            self .test_file_path =file_path 
            X_test ,label =load_data (file_path )
            self .test_file_label .setPlainText (f"File: {file_path}\nLabel: {label}")


    def display_results (self ,classifier_name ,confusion_matrix ,feature_importances ,test_confusion_matrix ,info_text ,labels ):


        for i in range (self .right_tab_widget .count ()):
            if self .right_tab_widget .tabText (i )==classifier_name :

                results_widget =self .right_tab_widget .widget (i )
                break 
        else :

            results_widget =ResultsWidget (classifier_name )
            self .right_tab_widget .addTab (results_widget ,classifier_name )

        if confusion_matrix is not None :
            results_widget .update_plots (confusion_matrix ,feature_importances ,None ,labels )


        if test_confusion_matrix is not None :
            results_widget .update_plots (None ,None ,test_confusion_matrix ,labels )

        results_widget .update_info (info_text )

if __name__ =="__main__":
    app =QApplication (sys .argv )
    app .setStyle (QStyleFactory .create ('Fusion'))

    palette =QPalette ()
    palette .setColor (QPalette .ColorRole .Window ,QColor (53 ,53 ,53 ))
    palette .setColor (QPalette .ColorRole .WindowText ,Qt .GlobalColor .white )
    palette .setColor (QPalette .ColorRole .Base ,QColor (25 ,25 ,25 ))
    palette .setColor (QPalette .ColorRole .AlternateBase ,QColor (53 ,53 ,53 ))
    palette .setColor (QPalette .ColorRole .ToolTipBase ,Qt .GlobalColor .white )
    palette .setColor (QPalette .ColorRole .ToolTipText ,Qt .GlobalColor .white )
    palette .setColor (QPalette .ColorRole .Text ,Qt .GlobalColor .white )
    palette .setColor (QPalette .ColorRole .Button ,QColor (53 ,53 ,53 ))
    palette .setColor (QPalette .ColorRole .ButtonText ,Qt .GlobalColor .white )
    palette .setColor (QPalette .ColorRole .BrightText ,Qt .GlobalColor .red )
    palette .setColor (QPalette .ColorRole .Link ,QColor (42 ,130 ,218 ))
    palette .setColor (QPalette .ColorRole .Highlight ,QColor (42 ,130 ,218 ))
    palette .setColor (QPalette .ColorRole .HighlightedText ,Qt .GlobalColor .black )
    app .setPalette (palette )
    window =MainWindow ()
    window .showMaximized ()
    sys .exit (app .exec ())


from __future__ import absolute_import 

import os 

from pkg_resources import packaging 


def check (cmd ,mf ):
    m =mf .findNode ("matplotlib")
    if m is None or m .filename is None :
        return None 




    from matplotlib import __version__ as VER 

    if cmd .matplotlib_backends :
        backends ={}
        for backend in cmd .matplotlib_backends :
            if backend =="-":
                pass 

            elif backend =="*":
                mf .import_hook ("matplotlib.backends",m ,["*"])

            else :
                mf .import_hook ("matplotlib.backends.backend_%s"%(backend ,),m )

    else :
        backends ={"packages":["matplotlib"]}

    result ={
    "resources":[os .path .join (os .path .dirname (m .filename ),"mpl-data")],
    }
    if packaging .version .parse (VER )<packaging .version .parse ("3.1"):
        result .update ({"prescripts":["py2app.recipes.matplotlib_prescript"]})

    result .update (backends )
    return result 



import sys 
from PySide6 import QtWidgets ,QtCore ,QtGui 
from uncertainties import ufloat 
from PySide6 .QtGui import QPalette ,QColor ,QFont 
from PySide6 .QtCore import Qt 

class MainWindow (QtWidgets .QMainWindow ):
    def __init__ (self ,*args ,**kwargs ):
        super (MainWindow ,self ).__init__ (*args ,**kwargs )


        self .setWindowTitle ("SD Nanopore Size Calculator")
        self .setGeometry (400 ,150 ,700 ,400 )


        self .main_widget =QtWidgets .QWidget ()

        self .setCentralWidget (self .main_widget )

        self .layout =QtWidgets .QGridLayout ()
        self .main_widget .setLayout (self .layout )


        self .title =QtWidgets .QLabel ("SD Nanopore Size Calculator\nshankar.dutt@anu.edu.au",self )
        self .title .setAlignment (QtCore .Qt .AlignmentFlag .AlignCenter )
        self .title .setFont (QtGui .QFont ('Arial',20 ))



        self .fields =["Conductance (nS)","Conductivity (S/m)","Pore Length (nm)"]
        self .inputs ={}
        for i ,field in enumerate (self .fields ):
            self .inputs [field ]=[QtWidgets .QLineEdit (),QtWidgets .QLineEdit ()]
            self .inputs [field ][0 ].setMinimumHeight (30 )
            self .inputs [field ][0 ].setFont (QtGui .QFont ('Arial',14 ))
            self .inputs [field ][1 ].setMinimumHeight (30 )
            self .inputs [field ][1 ].setFont (QtGui .QFont ('Arial',14 ))

            label =QtWidgets .QLabel (field )

            plus_minus =QtWidgets .QLabel ("±")


            self .layout .addWidget (label ,i +1 ,0 )
            self .layout .addWidget (self .inputs [field ][0 ],i +1 ,1 )
            self .layout .addWidget (plus_minus ,i +1 ,2 )
            self .layout .addWidget (self .inputs [field ][1 ],i +1 ,3 )


        self .calculate_button =QtWidgets .QPushButton ("Calculate")

        self .calculate_button .setFixedSize (200 ,50 )
        self .calculate_button .clicked .connect (self .calculate )


        self .result_label =QtWidgets .QLabel ("Diameter (nm): ")
        self .result =QtWidgets .QLabel ()




        self .layout .addWidget (self .title ,0 ,0 ,1 ,4 )
        self .layout .addWidget (self .calculate_button ,5 ,0 ,1 ,4 ,alignment =QtCore .Qt .AlignmentFlag .AlignCenter )
        self .layout .addWidget (self .result_label ,6 ,0 ,1 ,2 )
        self .layout .addWidget (self .result ,6 ,2 ,1 ,2 )

    def calculate (self ):
        from math import pi 

        try :
            G =ufloat (float (self .inputs ["Conductance (nS)"][0 ].text ()),float (self .inputs ["Conductance (nS)"][1 ].text ()))
            S =ufloat (float (self .inputs ["Conductivity (S/m)"][0 ].text ()),float (self .inputs ["Conductivity (S/m)"][1 ].text ()))
            L =ufloat (float (self .inputs ["Pore Length (nm)"][0 ].text ()),float (self .inputs ["Pore Length (nm)"][1 ].text ()))

            D =G /(2 *S )*(1 +(1 +16 *S *L /pi /G )**0.5 )
            self .result .setText (str (D ))

        except ValueError :
            error_dialog =QtWidgets .QMessageBox ()
            error_dialog .setIcon (QtWidgets .QMessageBox .Icon .Critical )
            error_dialog .setWindowTitle ("Input Error")
            error_dialog .setText ("Invalid input, please check your values.")
            error_dialog .exec ()

def main ():
    app =QtWidgets .QApplication (sys .argv )
    app .setStyle (QtWidgets .QStyleFactory .create ('Fusion'))


    palette =QPalette ()
    palette .setColor (QPalette .ColorRole .Window ,QColor (53 ,53 ,53 ))
    palette .setColor (QPalette .ColorRole .WindowText ,Qt .GlobalColor .white )
    palette .setColor (QPalette .ColorRole .Base ,QColor (25 ,25 ,25 ))
    palette .setColor (QPalette .ColorRole .AlternateBase ,QColor (53 ,53 ,53 ))
    palette .setColor (QPalette .ColorRole .ToolTipBase ,Qt .GlobalColor .white )
    palette .setColor (QPalette .ColorRole .ToolTipText ,Qt .GlobalColor .white )
    palette .setColor (QPalette .ColorRole .Text ,Qt .GlobalColor .white )
    palette .setColor (QPalette .ColorRole .Button ,QColor (53 ,53 ,53 ))
    palette .setColor (QPalette .ColorRole .ButtonText ,Qt .GlobalColor .white )
    palette .setColor (QPalette .ColorRole .BrightText ,Qt .GlobalColor .red )
    palette .setColor (QPalette .ColorRole .Link ,QColor (42 ,130 ,218 ))
    palette .setColor (QPalette .ColorRole .Highlight ,QColor (42 ,130 ,218 ))
    palette .setColor (QPalette .ColorRole .HighlightedText ,Qt .GlobalColor .black )
    app .setPalette (palette )
    main_window =MainWindow ()
    main_window .show ()

    sys .exit (app .exec ())

if __name__ =="__main__":
    main ()



import logging 
import os 
import sys 
import subprocess 
import warnings 
import json 
import datetime 

logging .basicConfig (level =logging .INFO )
logger =logging .getLogger (__name__ )

def check_and_install_nvidia_packages ():
    flag_file =os .path .join (os .path .dirname (__file__ ),'.nvidia_setup_complete')

    if os .path .exists (flag_file ):
        logger .info ("NVIDIA setup has already been completed. Skipping check and installation.")
        return 

    logger .info ("Checking for NVIDIA GPU...")
    try :
        import pynvml 
        pynvml .nvmlInit ()
        gpu_count =pynvml .nvmlDeviceGetCount ()
        if gpu_count >0 and (os .name =='nt'or sys .platform .startswith ('linux')):
            logger .info (f"Found {gpu_count} NVIDIA GPU(s).")
            try :
                driver_version =pynvml .nvmlSystemGetDriverVersion ()
                if isinstance (driver_version ,bytes ):
                    driver_version =driver_version .decode ('utf-8')
                logger .info (f"NVIDIA driver version: {driver_version}")
            except pynvml .NVMLError as e :
                logger .warning (f"Could not get driver version: {e}")

            logger .info ("Attempting to install NVIDIA packages...")
            try :
                subprocess .check_call ([
                sys .executable ,"-m","pip","install",
                "--extra-index-url=https://pypi.nvidia.com",
                "cudf-cu12==24.6.*","dask-cudf-cu12==24.6.*","cuml-cu12==24.6.*"
                ])
                subprocess .check_call ([
                sys .executable ,"-m","pip","install",
                "tensorrt"
                ])
                logger .info ("NVIDIA packages installed successfully.")
            except subprocess .CalledProcessError as e :
                logger .error (f"Failed to install NVIDIA packages: {e}")
                logger .info ("You may need to install these packages manually.")
        else :
            logger .info ("No NVIDIA GPU detected or not on Windows/Linux. Skipping NVIDIA package installation.")
    except ImportError :
        logger .warning ("pynvml is not installed. Skipping NVIDIA package check and installation.")
    except pynvml .NVMLError_DriverNotLoaded :
        logger .error ("NVIDIA driver is not loaded. Please check your NVIDIA driver installation.")
    except pynvml .NVMLError_LibraryNotFound :
        logger .error ("NVIDIA Management Library (NVML) not found. This might indicate an issue with your NVIDIA driver installation.")
    except pynvml .NVMLError as e :
        logger .error (f"An NVML error occurred: {e}")
        logger .info ("This might indicate a mismatch between your NVIDIA driver and CUDA versions.")
        logger .info ("Please ensure your NVIDIA driver and CUDA toolkit are compatible and up to date.")
    except Exception as e :
        logger .error (f"An unexpected error occurred while checking/installing NVIDIA packages: {e}")

    logger .info ("Continuing with package initialization...")


    with open (flag_file ,'w')as f :
        json .dump ({'setup_completed':True ,'timestamp':str (datetime .datetime .now ())},f )


check_and_install_nvidia_packages ()


from .nanosense import *

warnings .filterwarnings ("ignore",category =DeprecationWarning )

__version__ ="1.0.2"


"""
Automatic compilation of CoreData model files
"""
import os 

from py2app .decorators import converts 
from py2app .util import mapc ,momc 


@converts (suffix =".xcdatamodel")
def convert_datamodel (source ,destination ,dry_run =0 ):
    destination =os .path .splitext (destination )[0 ]+".mom"

    if dry_run :
        return 

    momc (source ,destination )


@converts (suffix =".xcmappingmodel")
def convert_mappingmodel (source ,destination ,dry_run =0 ):
    destination =destination [:-4 ]+".cdm"

    if dry_run :
        return 

    mapc (source ,destination )



""" Add Apple's additional packages to sys.path """


def add_system_python_extras ():
    import site 
    import sys 

    ver ="%s.%s"%(sys .version_info [:2 ])

    site .addsitedir (
    "/System/Library/Frameworks/Python.framework/Versions/"
    "%s/Extras/lib/python"%(ver ,)
    )


add_system_python_extras ()



import sys 
from PySide6 .QtWidgets import (QApplication ,QWidget ,QVBoxLayout ,QHBoxLayout ,
QPushButton ,QGroupBox ,QLabel ,QFileDialog ,
QScrollArea ,QMainWindow ,QSizePolicy ,QSplitter ,
QRadioButton ,QVBoxLayout ,QLineEdit ,QCheckBox ,QDoubleSpinBox ,QTabWidget ,
QTableWidget ,QTableWidgetItem ,QSpinBox ,QStyleFactory )
from PySide6 .QtCore import Qt ,QTimer ,QSize 
from PySide6 .QtGui import QFont ,QPalette ,QColor 
import numpy as np 
from matplotlib .backends .backend_qt5agg import FigureCanvasQTAgg as FigureCanvas 
from matplotlib .backends .backend_qt5agg import NavigationToolbar2QT 
from matplotlib .figure import Figure 

class MplCanvas (FigureCanvas ):
    def __init__ (self ,parent =None ,width =5 ,height =4 ,dpi =100 ):
        fig =Figure (figsize =(width ,height ),dpi =dpi )
        self .axes =fig .add_subplot (111 )
        super (MplCanvas ,self ).__init__ (fig )

class ScatterCanvas (FigureCanvas ):
    def __init__ (self ,parent =None ,width =5 ,height =4 ,dpi =100 ):
        fig =Figure (figsize =(width ,height ),dpi =dpi )
        self .axes =fig .add_subplot (111 )
        super (ScatterCanvas ,self ).__init__ (fig )

class SDAnalysisApp (QMainWindow ):
    def __init__ (self ):
        super ().__init__ ()
        self .setWindowTitle ('SD Event Analysis App')
        self .setGeometry (100 ,100 ,1200 ,800 )

        self .central_widget =QWidget ()
        self .setCentralWidget (self .central_widget )
        self .main_splitter =QSplitter (Qt .Orientation .Horizontal )


        self .setup_left_panel ()


        self .setup_right_panel ()


        self .central_layout =QHBoxLayout (self .central_widget )
        self .central_layout .addWidget (self .main_splitter )

        self .data =None 
        self .events_data ={}
        self .classification_to_event_ids ={}
        self .selected_event_ids =set ()


    def setup_left_panel (self ):
        self .top_group =QGroupBox ()
        self .top_group_layout =QVBoxLayout (self .top_group )


        self .configure_top_group ()


        self .event_categories_group_box ,self .event_categories_layout =self .setup_scroll_area ("Event Categories")
        self .event_classification_group_box ,self .event_classification_layout =self .setup_scroll_area ("Event Classifications")


        self .select_all_button =QPushButton ("Select All")
        self .select_all_button .clicked .connect (self .select_all_classifications )


        self .left_splitter =QSplitter (Qt .Orientation .Vertical )
        self .left_splitter .addWidget (self .top_group )
        self .left_splitter .addWidget (self .event_categories_group_box )
        self .left_splitter .addWidget (self .event_classification_group_box )
        self .left_splitter .addWidget (self .select_all_button )
        self .left_splitter .setSizes ([10 ,300 ,300 ,50 ])
        self .main_splitter .addWidget (self .left_splitter )

    def configure_top_group (self ):
        self .title_label =QLabel ('SD Event Analysis App')
        self .title_label .setFont (QFont ('Arial',23 ,QFont .Weight .Bold ))
        self .title_label .setAlignment (Qt .AlignmentFlag .AlignCenter )
        self .subtitle_label =QLabel ('shankar.dutt@anu.edu.au')
        self .subtitle_label .setFont (QFont ('Arial',15 ,QFont .Weight .Bold ))
        self .subtitle_label .setAlignment (Qt .AlignmentFlag .AlignCenter )
        self .top_group_layout .addWidget (self .title_label )
        self .top_group_layout .addWidget (self .subtitle_label )

        self .file_button =QPushButton ('Select File')
        self .file_button .clicked .connect (self .load_file )
        self .top_group_layout .addWidget (self .file_button )


        self .threshold_container =QWidget ()
        self .threshold_container_layout =QHBoxLayout (self .threshold_container )

        self .threshold_label =QLabel ("Similarity Threshold: ")
        self .threshold_container_layout .addWidget (self .threshold_label )

        self .threshold_input =QDoubleSpinBox ()
        self .threshold_input .setSuffix (" %")
        self .threshold_input .setRange (0 ,100 )
        self .threshold_input .setValue (85 )
        self .threshold_input .setSingleStep (1 )


        self .threshold_timer =QTimer (self )
        self .threshold_timer .setSingleShot (True )
        self .threshold_timer .timeout .connect (self .on_threshold_changed )
        self .threshold_input .valueChanged .connect (self .start_threshold_timer )

        self .threshold_container_layout .addWidget (self .threshold_input )

        self .reclassify_checkbox =QCheckBox ("Reclassify the event categories based on threshold")
        self .reclassify_checkbox .stateChanged .connect (self .on_threshold_changed )
        self .top_group_layout .addWidget (self .threshold_container )
        self .top_group_layout .addWidget (self .reclassify_checkbox )

    def start_threshold_timer (self ):
        self .threshold_timer .start (500 )

    def setup_right_panel (self ):
        self .right_panel =QWidget ()
        self .right_splitter =QSplitter (Qt .Orientation .Vertical )


        self .tabs =QTabWidget ()
        self .histograms_tab =QWidget ()
        self .scatter_plots_tab =QWidget ()
        self .tabs .addTab (self .histograms_tab ,"Histograms")
        self .tabs .addTab (self .scatter_plots_tab ,"Scatter Plots")


        self .setup_histograms_tab ()


        self .setup_scatter_plots_tab ()

        self .right_splitter .addWidget (self .tabs )


        self .histograms_group =QGroupBox ("Histograms")
        self .histograms_horizontal_layout =QHBoxLayout (self .histograms_group )



        self .bottom_right_splitter =QSplitter (Qt .Orientation .Horizontal )


        self .event_plots_group =QGroupBox ("Event Plots")
        self .event_plots_layout =QVBoxLayout (self .event_plots_group )
        self .event_plot_canvas =FigureCanvas (Figure (figsize =(5 ,3 )))
        self .event_plots_layout .addWidget (self .event_plot_canvas )
        self .event_plot_toolbar =NavigationToolbar2QT (self .event_plot_canvas ,self .event_plots_group )
        self .event_plot_toolbar .setIconSize (QSize (16 ,16 ))
        self .event_plots_layout .addWidget (self .event_plot_toolbar )


        self .event_navigation_layout =QHBoxLayout ()
        self .prev_button =QPushButton ("Previous")
        self .event_navigation_layout .addWidget (self .prev_button )

        self .next_button =QPushButton ("Next")
        self .event_navigation_layout .addWidget (self .next_button )
        self .event_plots_layout .addLayout (self .event_navigation_layout )
        self .next_button .clicked .connect (self .next_event )
        self .prev_button .clicked .connect (self .previous_event )


        self .bottom_right_splitter .addWidget (self .event_plots_group )


        self .table_group =QGroupBox ("Event Information")
        self .table_layout =QVBoxLayout (self .table_group )
        self .event_info_table =QTableWidget (10 ,3 )
        self .event_info_table .setHorizontalHeaderLabels (['Type','Value','Description'])
        self .table_layout .addWidget (self .event_info_table )
        self .bottom_right_splitter .addWidget (self .table_group )

        self .right_splitter .addWidget (self .bottom_right_splitter )
        self .main_splitter .addWidget (self .right_splitter )


        self .main_splitter .setSizes ([300 ,800 ])
        self .right_splitter .setSizes ([400 ,350 ])
        self .bottom_right_splitter .setSizes ([400 ,400 ])

    def setup_histograms_tab (self ):
        self .histograms_layout =QVBoxLayout (self .histograms_tab )
        self .histograms_group =QGroupBox ("Histograms")
        self .histograms_horizontal_layout =QHBoxLayout (self .histograms_group )


        self .all_events_layout =QVBoxLayout ()
        self .all_events_histogram_canvas =FigureCanvas (Figure (figsize =(5 ,3 )))
        self .all_events_layout .addWidget (self .all_events_histogram_canvas )
        self .all_events_histogram_toolbar =NavigationToolbar2QT (self .all_events_histogram_canvas ,self .histograms_group )
        self .all_events_histogram_toolbar .setIconSize (QSize (16 ,16 ))
        self .all_events_layout .addWidget (self .all_events_histogram_toolbar )
        self .histograms_horizontal_layout .addLayout (self .all_events_layout )


        self .selected_classifications_layout =QVBoxLayout ()
        self .selected_classifications_histogram_canvas =FigureCanvas (Figure (figsize =(5 ,3 )))
        self .selected_classifications_layout .addWidget (self .selected_classifications_histogram_canvas )
        self .selected_classifications_histogram_toolbar =NavigationToolbar2QT (self .selected_classifications_histogram_canvas ,self .histograms_group )
        self .selected_classifications_histogram_toolbar .setIconSize (QSize (16 ,16 ))
        self .selected_classifications_layout .addWidget (self .selected_classifications_histogram_toolbar )
        self .histograms_horizontal_layout .addLayout (self .selected_classifications_layout )

        self .histograms_layout .addWidget (self .histograms_group )

    def setup_scatter_plots_tab (self ):
        self .scatter_plots_layout =QVBoxLayout (self .scatter_plots_tab )
        self .scatter_plots_group =QGroupBox ("Scatter Plots")
        self .scatter_plots_horizontal_layout =QHBoxLayout (self .scatter_plots_group )


        self .all_events_scatter_layout =QVBoxLayout ()
        self .all_events_scatter_canvas =ScatterCanvas (self ,width =5 ,height =3 )
        self .all_events_scatter_layout .addWidget (self .all_events_scatter_canvas )
        self .all_events_scatter_toolbar =NavigationToolbar2QT (self .all_events_scatter_canvas ,self .scatter_plots_group )
        self .all_events_scatter_toolbar .setIconSize (QSize (16 ,16 ))
        self .all_events_scatter_layout .addWidget (self .all_events_scatter_toolbar )
        self .scatter_plots_horizontal_layout .addLayout (self .all_events_scatter_layout )


        self .selected_classifications_scatter_layout =QVBoxLayout ()
        self .selected_classifications_scatter_canvas =ScatterCanvas (self ,width =5 ,height =3 )
        self .selected_classifications_scatter_layout .addWidget (self .selected_classifications_scatter_canvas )
        self .selected_classifications_scatter_toolbar =NavigationToolbar2QT (self .selected_classifications_scatter_canvas ,self .scatter_plots_group )
        self .selected_classifications_scatter_toolbar .setIconSize (QSize (16 ,16 ))
        self .selected_classifications_scatter_layout .addWidget (self .selected_classifications_scatter_toolbar )
        self .scatter_plots_horizontal_layout .addLayout (self .selected_classifications_scatter_layout )

        self .scatter_plots_layout .addWidget (self .scatter_plots_group )


    def setup_scroll_area (self ,title ):
        scroll_area =QScrollArea ()
        scroll_area .setWidgetResizable (True )

        container_widget =QWidget ()
        layout =QVBoxLayout (container_widget )

        scroll_area .setWidget (container_widget )

        group_box =QGroupBox (title )
        group_layout =QVBoxLayout (group_box )
        group_layout .addWidget (scroll_area )

        return group_box ,layout 

    def on_threshold_changed (self ):

        self .prepare_and_display_event_data ()


    def load_file (self ):
        file_name ,_ =QFileDialog .getOpenFileName (self ,'Open NPZ File','','NPZ Files (*event_fitting.npz)')
        if file_name :
            self .data =np .load (file_name ,allow_pickle =True )
            self .prepare_and_display_event_data ()
































    def prepare_and_display_event_data (self ):
        if self .data is not None :
            self .events_data .clear ()
            self .clear_layout (self .event_categories_layout )
            self .clear_layout (self .event_classification_layout )
            self .plot_all_events_histogram ()

            for key in self .data .files :
                if 'SEGMENT_INFO'in key and 'number_of_segments'in key :
                    event_id =int (key .split ('_')[2 ])
                    mean_diffs_key =f'SEGMENT_INFO_{event_id}_segment_mean_diffs'
                    segment_widths_key =f'SEGMENT_INFO_{event_id}_segment_widths_time'
                    if mean_diffs_key in self .data and segment_widths_key in self .data :
                        mean_diffs =self .data [mean_diffs_key ]
                        segment_widths =self .data [segment_widths_key ]
                        classification ,new_segment_count =self .classify_event (mean_diffs )
                        self .events_data .setdefault (new_segment_count ,[]).append ((event_id ,mean_diffs ,segment_widths ,classification ))
                    else :
                        new_segment_count =int (self .data [key ][0 ])
                        self .events_data .setdefault (new_segment_count ,[]).append ((event_id ,[],[],''))

            for num_segments in sorted (self .events_data ):
                events_data =self .events_data [num_segments ]
                event_ids =[event_data [0 ]for event_data in events_data ]
                radio_button =QRadioButton (f"{num_segments} segments ({len(event_ids)} events)")
                radio_button .segment_number =num_segments 
                radio_button .toggled .connect (self .on_radio_button_toggled )
                self .event_categories_layout .addWidget (radio_button )

    def on_radio_button_toggled (self ):
        radio_button =self .sender ()
        if radio_button .isChecked ():
            self .update_classification_group (radio_button .segment_number )

    def update_classification_group (self ,segment_number ):
        self .clear_layout (self .event_classification_layout )
        classification_counts ={}
        self .classification_checkboxes =[]
        self .classification_to_event_ids .clear ()

        for event_data in self .events_data .get (segment_number ,[]):
            event_id =event_data [0 ]
            mean_diffs_key =f'SEGMENT_INFO_{event_id}_segment_mean_diffs'
            if mean_diffs_key in self .data :
                mean_diffs =self .data [mean_diffs_key ]
                classification ,_ =self .classify_event (mean_diffs )
                classification_counts [classification ]=classification_counts .get (classification ,0 )+1 

                if classification not in self .classification_to_event_ids :
                    self .classification_to_event_ids [classification ]=[]
                self .classification_to_event_ids [classification ].append (event_id )

        for classification ,count in sorted (classification_counts .items (),key =lambda x :x [0 ]):
            checkbox =QCheckBox (f"Category {classification} ({count} events)")

            checkbox .classification =classification 
            checkbox .stateChanged .connect (self .on_checkbox_state_changed )
            self .event_classification_layout .addWidget (checkbox )
            self .classification_checkboxes .append (checkbox )


        self .select_all_button .clicked .disconnect ()
        self .select_all_button .clicked .connect (self .select_all_classifications )

    def classify_event (self ,mean_diffs ):
        threshold_ratio =self .threshold_input .value ()/100.0 


        abs_mean_diffs =np .abs (mean_diffs )


        if len (abs_mean_diffs )<=1 :
            return "1",1 


        sorted_indices =sorted (range (len (abs_mean_diffs )),key =lambda k :abs_mean_diffs [k ])
        sorted_abs_mean_diffs =[abs_mean_diffs [i ]for i in sorted_indices ]


        classifications =[0 ]*len (abs_mean_diffs )
        unique_classes =[]

        for i ,idx in enumerate (sorted_indices ):
            if i ==0 or sorted_abs_mean_diffs [i ]/sorted_abs_mean_diffs [i -1 ]>threshold_ratio :
                new_class =str (len (unique_classes )+1 )
                unique_classes .append (new_class )
                classifications [idx ]=new_class 
            else :
                classifications [idx ]=classifications [sorted_indices [i -1 ]]


        if self .reclassify_checkbox .isChecked ()and len (classifications )>1 :

            new_classifications =[classifications [0 ]]
            for i in range (1 ,len (classifications )):

                if classifications [i ]==classifications [i -1 ]:
                    continue 
                new_classifications .append (classifications [i ])
            classifications =new_classifications 


        classification_string =''.join (classifications )


        new_segment_count =len (set (classification_string ))


        return classification_string ,new_segment_count 

    def plot_selected_events_scatter (self ):
        all_mean_diffs =[]
        all_segment_widths =[]
        for event_id in self .selected_event_ids :
            for segment_count ,events in self .events_data .items ():
                for event_data in events :
                    if event_data [0 ]==event_id :
                        _ ,mean_diffs ,segment_widths ,_ =event_data 
                        all_mean_diffs .extend (mean_diffs )
                        all_segment_widths .extend (segment_widths )
                        break 

        if len (all_mean_diffs )>0 and len (all_segment_widths )>0 :

            self .selected_classifications_scatter_canvas .figure .clear ()

            ax =self .selected_classifications_scatter_canvas .figure .subplots ()
            ax .scatter (np .log (np .array (all_segment_widths )*1e3 ),all_mean_diffs )
            ax .set_title ('Segment Mean Diffs vs log(dt (ms)) for Selected Events')
            ax .set_xlabel ('log(Δt (ms))')
            ax .set_ylabel ('Mean Diff')
            self .selected_classifications_scatter_canvas .figure .tight_layout ()
            self .selected_classifications_scatter_canvas .draw ()














































    def clear_layout (self ,layout ):
        while layout .count ():
            child =layout .takeAt (0 )
            if child .widget ():
                child .widget ().deleteLater ()

    def select_all_classifications (self ):
        try :
            if any (not checkbox .isChecked ()for checkbox in self .classification_checkboxes ):
                for checkbox in self .classification_checkboxes :
                    checkbox .setChecked (True )
            else :
                for checkbox in self .classification_checkboxes :
                    checkbox .setChecked (False )


            self .select_all_button .setText ("Unselect All"if any (checkbox .isChecked ()for checkbox in self .classification_checkboxes )else "Select All")
        except :
            pass 


    def plot_all_events_histogram (self ):
        max_mean_diffs =[]
        for key in self .data .files :
            if 'segment_mean_diffs'in key :
                mean_diffs =self .data [key ]
                max_mean_diffs .append (np .max (mean_diffs ))


        self .all_events_histogram_canvas .figure .clear ()

        ax =self .all_events_histogram_canvas .figure .subplots ()


        _ ,bins_auto =np .histogram (max_mean_diffs ,bins ='auto')
        num_bins_auto =len (bins_auto )-1 
        doubled_num_bins =num_bins_auto *2 


        new_bins =np .linspace (bins_auto [0 ],bins_auto [-1 ],doubled_num_bins +1 )

        ax .hist (max_mean_diffs ,bins =new_bins )
        ax .set_title ('Max Segment Mean Diffs for All Events')
        ax .set_xlabel ('Max Mean Diff')
        ax .set_ylabel ('Frequency')
        self .all_events_histogram_canvas .figure .tight_layout ()
        self .all_events_histogram_canvas .draw ()

        self .plot_all_events_scatter ()

    def on_checkbox_state_changed (self ):
        self .selected_event_ids .clear ()

        for checkbox in self .classification_checkboxes :
            if checkbox .isChecked ():

                event_ids =self .classification_to_event_ids .get (checkbox .classification ,[])
                self .selected_event_ids .update (event_ids )


        self .current_event_index =0 
        if self .selected_event_ids :
            first_event_id =next (iter (sorted (self .selected_event_ids )))
            self .plot_event_data (first_event_id )
            self .display_segment_info (first_event_id )


        self .plot_selected_events_histogram ()

    def setup_navigation_buttons (self ):
        self .current_event_index =0 
        self .previous_button .clicked .connect (self .previous_event )
        self .next_button .clicked .connect (self .next_event )

    def previous_event (self ):
        if self .selected_event_ids and self .current_event_index >0 :
            self .current_event_index -=1 
            event_id =sorted (self .selected_event_ids )[self .current_event_index ]
            self .plot_event_data (event_id )
            self .display_segment_info (event_id )

    def next_event (self ):
        if self .selected_event_ids and self .current_event_index <len (self .selected_event_ids )-1 :
            self .current_event_index +=1 
            event_id =sorted (self .selected_event_ids )[self .current_event_index ]
            self .plot_event_data (event_id )
            self .display_segment_info (event_id )

    def plot_selected_events_histogram (self ):
        all_mean_diffs =[]
        for event_id in self .selected_event_ids :
            key =f'SEGMENT_INFO_{event_id}_segment_mean_diffs'
            if key in self .data :
                mean_diffs =self .data [key ]
                all_mean_diffs .extend (mean_diffs )

        if len (all_mean_diffs )>1 :

            self .selected_classifications_histogram_canvas .figure .clear ()

            ax =self .selected_classifications_histogram_canvas .figure .subplots ()


            num_bins ='auto'


            ax .hist (all_mean_diffs ,bins =num_bins ,edgecolor ='black')

            ax .set_title ('Segment Mean Diffs for Selected Events')
            ax .set_xlabel ('Mean Diff')
            ax .set_ylabel ('Frequency')
            self .selected_classifications_histogram_canvas .figure .tight_layout ()
            self .selected_classifications_histogram_canvas .draw ()


        self .plot_selected_events_scatter ()

    def plot_all_events_scatter (self ):
        max_mean_diffs =[]
        event_widths =[]
        for key in self .data .files :
            if 'segment_mean_diffs'in key :
                mean_diffs =self .data [key ]
                max_mean_diffs .append (np .max (mean_diffs ))

            if 'event_width'in key :
                event_widths .append (self .data [key ])

        if len (max_mean_diffs )>0 and len (event_widths )>0 :

            self .all_events_scatter_canvas .figure .clear ()

            ax =self .all_events_scatter_canvas .figure .subplots ()
            ax .scatter (np .log (np .array (event_widths )*1e3 ),max_mean_diffs )
            ax .set_title ('Max Segment Mean Diffs vs log(Event Width) for All Events')
            ax .set_xlabel ('log(Δt (ms))')
            ax .set_ylabel ('ΔI')
            self .all_events_scatter_canvas .figure .tight_layout ()
            self .all_events_scatter_canvas .draw ()

    def plot_event_data (self ,event_id ):

        x_values =self .data [f'EVENT_DATA_{event_id}_part_0']
        y_values_event =self .data [f'EVENT_DATA_{event_id}_part_1']
        y_values_fit =self .data [f'EVENT_DATA_{event_id}_part_3']


        self .event_plot_canvas .figure .clear ()


        ax =self .event_plot_canvas .figure .subplots ()
        ax .plot (x_values ,y_values_event ,label ='Event Data')
        ax .plot (x_values ,y_values_fit ,label ='Fit Data',linestyle ='--')


        ax .legend ()
        ax .set_title (f'Event {event_id} Data')
        ax .set_xlabel ('Time')
        ax .set_ylabel ('Data')

        self .event_plot_canvas .figure .tight_layout ()

        self .event_plot_canvas .draw ()



    def display_segment_info (self ,event_id ):
        segment_info_keys =[
        f'SEGMENT_INFO_{event_id}_number_of_segments',
        f'SEGMENT_INFO_{event_id}_segment_mean_diffs',
        f'SEGMENT_INFO_{event_id}_segment_widths_time'
        ]


        number_of_segments =self .data [segment_info_keys [0 ]][0 ]
        segment_mean_diffs =self .data [segment_info_keys [1 ]]
        segment_widths_time =self .data [segment_info_keys [2 ]]


        self .event_info_table .setRowCount (int (number_of_segments ))
        self .event_info_table .setColumnCount (3 )
        self .event_info_table .setHorizontalHeaderLabels (['Segment','Mean Diff','Width Time'])


        for i in range (int (number_of_segments )):
            self .event_info_table .setItem (i ,0 ,QTableWidgetItem (f"{i + 1}"))
            self .event_info_table .setItem (i ,1 ,QTableWidgetItem (f"{segment_mean_diffs[i]:.3g}"))
            self .event_info_table .setItem (i ,2 ,QTableWidgetItem (f"{segment_widths_time[i]:.3g}"))


        self .event_info_table .resizeColumnsToContents ()


        self .event_info_table .resizeRowsToContents ()


if __name__ =="__main__":
    app =QApplication (sys .argv )
    app .setStyle (QStyleFactory .create ('Fusion'))
    palette =QPalette ()
    palette .setColor (QPalette .ColorRole .Window ,QColor (53 ,53 ,53 ))
    palette .setColor (QPalette .ColorRole .WindowText ,Qt .GlobalColor .white )
    palette .setColor (QPalette .ColorRole .Base ,QColor (25 ,25 ,25 ))
    palette .setColor (QPalette .ColorRole .AlternateBase ,QColor (53 ,53 ,53 ))
    palette .setColor (QPalette .ColorRole .ToolTipBase ,Qt .GlobalColor .white )
    palette .setColor (QPalette .ColorRole .ToolTipText ,Qt .GlobalColor .white )
    palette .setColor (QPalette .ColorRole .Text ,Qt .GlobalColor .white )
    palette .setColor (QPalette .ColorRole .Button ,QColor (53 ,53 ,53 ))
    palette .setColor (QPalette .ColorRole .ButtonText ,Qt .GlobalColor .white )
    palette .setColor (QPalette .ColorRole .BrightText ,Qt .GlobalColor .red )
    palette .setColor (QPalette .ColorRole .Link ,QColor (42 ,130 ,218 ))
    palette .setColor (QPalette .ColorRole .Highlight ,QColor (42 ,130 ,218 ))
    palette .setColor (QPalette .ColorRole .HighlightedText ,Qt .GlobalColor .black )
    app .setPalette (palette )
    mainWin =SDAnalysisApp ()
    mainWin .showMaximized ()
    sys .exit (app .exec ())



from setuptools import setup, find_packages

setup(
    name="nanosense",
    version="1.0.2",  # Make sure this matches the version in pyproject.toml
    packages=find_packages(where="src"),
    package_dir={"": "src"},
    include_package_data=True,
    entry_points={
        'console_scripts': [
            'nanosense=nanosense.nanosense:main',
        ],
    },
)


import os 
import sys 

from macholib .util import in_system_path 
from modulegraph import modulegraph 


def has_filename_filter (module ):
    if isinstance (module ,modulegraph .MissingModule ):
        return True 
    if hasattr (modulegraph ,"InvalidRelativeImport")and isinstance (
    module ,modulegraph .InvalidRelativeImport 
    ):
        return True 
    return getattr (module ,"filename",None )is not None 


def not_stdlib_filter (module ,prefix =None ):
    """
    Return False if the module is located in the standard library
    """
    if prefix is None :
        prefix =sys .prefix 
    if module .filename is None :
        return True 

    prefix =os .path .join (os .path .realpath (prefix ),"")

    rp =os .path .realpath (module .filename )
    if rp .startswith (prefix ):
        rest =rp [len (prefix ):]
        if "/site-python/"in rest :
            return True 
        elif "/site-packages/"in rest :
            return True 
        else :
            return False 

    if os .path .exists (os .path .join (prefix ,".Python")):

        fn =os .path .join (
        prefix ,"lib","python%d.%d"%(sys .version_info [:2 ]),"orig-prefix.txt"
        )

        if os .path .exists (fn ):
            if sys .version_info [0 ]==2 :
                with open (fn ,"rU")as fp :
                    prefix =fp .read ().strip ()
            else :
                with open (fn ,"r")as fp :
                    prefix =fp .read ().strip ()

            if rp .startswith (prefix ):
                rest =rp [len (prefix ):]
                if "/site-python/"in rest :
                    return True 
                elif "/site-packages/"in rest :
                    return True 
                else :
                    return False 

    if hasattr (sys ,"base_prefix"):

        prefix =os .path .join (os .path .realpath (sys .base_prefix ),"")
        if rp .startswith (prefix ):
            rest =rp [len (prefix ):]
            if "/site-python/"in rest :
                return True 
            elif "/site-packages/"in rest :
                return True 
            else :
                return False 

    return True 


def not_system_filter (module ):
    """
    Return False if the module is located in a system directory
    """
    return not in_system_path (module .filename )


def bundle_or_dylib_filter (module ):
    """
    Return False if the module does not have a filetype attribute
    corresponding to a Mach-O bundle or dylib
    """
    return getattr (module ,"filetype",None )in ("bundle","dylib")



import re 
import sys 

cookie_re =re .compile (br"coding[:=]\s*([-\w.]+)")
if sys .version_info [0 ]==2 :
    default_encoding ="ascii"
else :
    default_encoding ="utf-8"


def guess_encoding (fp ):
    for _i in range (2 ):
        ln =fp .readline ()

        m =cookie_re .search (ln )
        if m is not None :
            return m .group (1 ).decode ("ascii")

    return default_encoding 


def _run ():
    global __file__ 
    import os 
    import site 

    sys .frozen ="macosx_app"
    base =os .environ ["RESOURCEPATH"]

    argv0 =os .path .basename (os .environ ["ARGVZERO"])
    script =SCRIPT_MAP .get (argv0 ,DEFAULT_SCRIPT )

    path =os .path .join (base ,script )
    sys .argv [0 ]=__file__ =path 
    if sys .version_info [0 ]==2 :
        with open (path ,"rU")as fp :
            source =fp .read ()+"\n"
    else :
        with open (path ,"rb")as fp :
            encoding =guess_encoding (fp )

        with open (path ,"r",encoding =encoding )as fp :
            source =fp .read ()+"\n"

        BOM =b"\xef\xbb\xbf".decode ("utf-8")
        if source .startswith (BOM ):
            source =source [1 :]

    exec (compile (source ,path ,"exec"),globals (),globals ())



import os 
import plistlib 
import shutil 
import sys 

from pkg_resources import resource_filename 

import py2app .bundletemplate 
from py2app .util import make_exec ,makedirs ,mergecopy ,mergetree ,skipscm 


def create_pluginbundle (
destdir ,
name ,
extension =".plugin",
module =py2app .bundletemplate ,
platform ="MacOS",
copy =mergecopy ,
mergetree =mergetree ,
condition =skipscm ,
plist =None ,
arch =None ,
):
    if plist is None :
        plist ={}

    kw =module .plist_template .infoPlistDict (
    plist .get ("CFBundleExecutable",name ),plist 
    )
    plugin =os .path .join (destdir ,kw ["CFBundleName"]+extension )
    if os .path .exists (plugin ):


        shutil .rmtree (plugin )
    contents =os .path .join (plugin ,"Contents")
    resources =os .path .join (contents ,"Resources")
    platdir =os .path .join (contents ,platform )
    dirs =[contents ,resources ,platdir ]
    plist ={}
    plist .update (kw )
    plistPath =os .path .join (contents ,"Info.plist")
    if os .path .exists (plistPath ):
        with open (plistPath ,"rb")as fp :
            if hasattr (plistlib ,"load"):
                contents =plistlib .load (fp )
            else :

                contents =plistlib .readPlist (fp )

            if plist !=contents :
                for d in dirs :
                    shutil .rmtree (d ,ignore_errors =True )
    for d in dirs :
        makedirs (d )

    with open (plistPath ,"wb")as fp :
        if hasattr (plistlib ,"dump"):
            plistlib .dump (plist ,fp )
        else :
            plistlib .writePlist (plist ,fp )
    srcmain =module .setup .main (arch =arch )
    if sys .version_info [0 ]==2 and isinstance (
    kw ["CFBundleExecutable"],unicode 
    ):
        destmain =os .path .join (platdir ,kw ["CFBundleExecutable"].encode ("utf-8"))
    else :
        destmain =os .path .join (platdir ,kw ["CFBundleExecutable"])
    with open (os .path .join (contents ,"PkgInfo"),"w")as fp :
        fp .write (kw ["CFBundlePackageType"]+kw ["CFBundleSignature"])
    copy (srcmain ,destmain )
    make_exec (destmain )
    mergetree (
    resource_filename (module .__name__ ,"lib"),
    resources ,
    condition =condition ,
    copyfn =copy ,
    )
    return plugin ,plist 


if __name__ =="__main__":
    create_pluginbundle ("build",sys .argv [1 ])



import sys 
from PySide6 .QtWidgets import QApplication ,QMainWindow ,QWidget ,QHBoxLayout ,QVBoxLayout ,QPushButton ,QCheckBox ,QListWidget ,QLabel ,QFileDialog ,QLineEdit ,QMessageBox ,QListWidgetItem ,QSplitter ,QGroupBox ,QFormLayout ,QComboBox ,QSpinBox ,QStyleFactory ,QTabWidget ,QScrollArea ,QGridLayout ,QComboBox ,QInputDialog 
from PySide6 .QtCore import Qt 
from PySide6 .QtGui import QPalette ,QColor 
import os 
import numpy as np 
from neo .rawio import AxonRawIO 
from scipy .ndimage import uniform_filter1d 
import matplotlib 
from matplotlib .backends .backend_qtagg import FigureCanvasQTAgg ,NavigationToolbar2QT 
from matplotlib .figure import Figure 
import h5py 
from scipy .signal import butter ,bessel ,cheby1 ,cheby2 ,ellip ,firwin ,lfilter ,sosfilt ,sosfilt_zi 
from PySide6 .QtWidgets import QProgressDialog 
from PySide6 .QtCore import Qt ,QSize 
from concurrent .futures import ThreadPoolExecutor 
matplotlib .use ('Qt5Agg')
import math 

class MplCanvas (FigureCanvasQTAgg ):
    def __init__ (self ,parent =None ,width =5 ,height =4 ,dpi =100 ):
        fig =Figure (figsize =(width ,height ),dpi =dpi )
        self .axes =fig .add_subplot (111 )
        fig .tight_layout ()
        super ().__init__ (fig )

def load_abf_file (file_path ):

    raw_io =AxonRawIO (filename =file_path )


    raw_io .parse_header ()


    channel_index =0 


    signal_size =raw_io .get_signal_size (block_index =0 ,seg_index =0 )



    data =raw_io .get_analogsignal_chunk (block_index =0 ,seg_index =0 ,i_start =0 ,i_stop =signal_size ,channel_indexes =[channel_index ])


    data =raw_io .rescale_signal_raw_to_float (data ,dtype ='float64',channel_indexes =[channel_index ]).flatten ()


    sampling_rate =raw_io .get_signal_sampling_rate ()


    time =np .arange (len (data ))/sampling_rate 

    return data ,sampling_rate ,time 

def load_hdf5_file (file_path ):
    with h5py .File (file_path ,'r')as f :
        selected_data =f ['selected_data'][()]
        sampling_rate =f .attrs ['sampling_rate']
        time =np .arange (len (selected_data ))/sampling_rate 
    return selected_data ,sampling_rate ,time 

def calculate_rolling_stats (data ,sampling_rate ,avg_window_size_in_ms ):
    avg_window_size_samples =int ((avg_window_size_in_ms /1000 )*sampling_rate )
    if avg_window_size_samples <1 :
        return data 
    return uniform_filter1d (data ,size =avg_window_size_samples )


class Settings :
    def __init__ (self ):
        self .calculate_avg =False 
        self .avg_window_size =0 
        self .use_threshold =False 
        self .threshold_value =0 
        self .threshold_percentage =75 
        self .use_segments =False 
        self .segments =[]

    def __eq__ (self ,other ):
        if not isinstance (other ,Settings ):
            return NotImplemented 
        return (self .calculate_avg ==other .calculate_avg and 
        self .avg_window_size ==other .avg_window_size and 
        self .use_threshold ==other .use_threshold and 
        self .threshold_value ==other .threshold_value and 
        self .threshold_percentage ==other .threshold_percentage and 
        self .use_segments ==other .use_segments and 
        self .segments ==other .segments )

class PlotData :
    def __init__ (self ,file_path ,data ,unfiltered_data ,sampling_rate ,time ):
        self .file_path =file_path 
        self .data =data 
        self .unfiltered_data =unfiltered_data 
        self .sampling_rate =sampling_rate 
        self .time =time 
        self .mean_value =np .mean (data )
        self .settings ={
        'calculate_avg':False ,
        'avg_window_size':10 ,
        'use_threshold':False ,
        'threshold_value':self .mean_value ,
        'threshold_percentage':75 ,
        'use_segments':False ,
        'segments':[]
        }
        self .original_settings =self .settings .copy ()



class SegmentWidget (QWidget ):
    def __init__ (self ,parent =None ):
        super ().__init__ (parent )
        layout =QHBoxLayout ()
        self .setLayout (layout )

        self .start_label =QLabel ("Start (s):")
        self .start_input =QLineEdit ()
        self .end_label =QLabel ("End (s):")
        self .end_input =QLineEdit ()
        self .add_button =QPushButton ("Add Segment")
        self .apply_all_checkbox =QCheckBox ("Apply to All Plots")

        layout .addWidget (self .start_label )
        layout .addWidget (self .start_input )
        layout .addWidget (self .end_label )
        layout .addWidget (self .end_input )
        layout .addWidget (self .add_button )
        layout .addWidget (self .apply_all_checkbox )


class MainWindow (QMainWindow ):
    def __init__ (self ):
        super ().__init__ ()
        self .setWindowTitle ("SD ABF File Plotter and Selector App")
        self .setGeometry (100 ,100 ,1200 ,800 )

        self .plot_data_dict ={}
        self .current_file_path =None 
        self .global_settings =Settings ()

        main_splitter =QSplitter (Qt .Orientation .Horizontal )

        left_widget =QWidget ()
        left_layout =QVBoxLayout ()
        left_widget .setLayout (left_layout )

        right_widget =QWidget ()
        right_layout =QVBoxLayout ()
        right_widget .setLayout (right_layout )

        main_splitter .addWidget (left_widget )
        main_splitter .addWidget (right_widget )
        main_splitter .setSizes ([300 ,900 ])

        self .setCentralWidget (main_splitter )


        self .app_name_label =QLabel ("SD ABF File Plotter and Selector App")
        self .app_name_label .setAlignment (Qt .AlignmentFlag .AlignCenter )
        self .app_name_label .setStyleSheet ("font-size: 22px; font-weight: bold;")
        self .email_label =QLabel ("shankar.dutt@anu.edu.au")
        self .email_label .setAlignment (Qt .AlignmentFlag .AlignCenter )

        self .select_folder_btn =QPushButton ("Select Folder")
        self .select_folder_btn .clicked .connect (self .select_folder )

        self .include_subfolders_chk =QCheckBox ("Include Subfolders")

        self .files_list_widget =QListWidget ()
        self .files_list_widget .setSelectionMode (QListWidget .SelectionMode .MultiSelection )

        self .folder_path_label =QLabel (" ")
        self .folder_path_label .setWordWrap (True )

        self .plot_multiple_files_chk =QCheckBox ("Plot multiple files")
        self .plot_multiple_files_chk .stateChanged .connect (self .toggle_plot_multiple_files )

        nth_element_layout =QHBoxLayout ()
        self .nth_element_label =QLabel ("Plot nth element:")
        self .nth_element_spinbox =QSpinBox ()
        self .nth_element_spinbox .setMinimum (1 )
        self .nth_element_spinbox .setMaximum (1000 )
        self .nth_element_spinbox .setValue (1 )
        nth_element_layout .addWidget (self .nth_element_label )
        nth_element_layout .addWidget (self .nth_element_spinbox )

        self .low_pass_filter_chk =QCheckBox ("Apply Low Pass Filter")
        self .low_pass_filter_chk .stateChanged .connect (self .toggle_low_pass_filter )

        self .filter_type_label =QLabel ("Filter Type:")
        self .filter_type_dropdown =QComboBox ()
        self .filter_type_dropdown .addItems (["Butterworth","Bessel","Chebyshev I","Chebyshev II","Elliptic","FIR","IIR"])
        self .filter_type_dropdown .setEnabled (False )

        self .cutoff_frequency_label =QLabel ("Cutoff Frequency (kHz):")
        self .cutoff_frequency_spinbox =QSpinBox ()
        self .cutoff_frequency_spinbox .setRange (1 ,1000 )
        self .cutoff_frequency_spinbox .setValue (10 )
        self .cutoff_frequency_spinbox .setEnabled (False )

        self .plot_btn =QPushButton ("Plot Selected File(s)")
        self .plot_btn .clicked .connect (self .plot_selected_files )

        left_layout .addWidget (self .app_name_label )
        left_layout .addWidget (self .email_label )
        left_layout .addWidget (self .select_folder_btn )
        left_layout .addWidget (self .include_subfolders_chk )
        left_layout .addWidget (self .files_list_widget )
        left_layout .addWidget (self .folder_path_label )
        left_layout .addWidget (self .plot_multiple_files_chk )
        left_layout .addLayout (nth_element_layout )
        left_layout .addWidget (self .low_pass_filter_chk )
        left_layout .addWidget (self .filter_type_label )
        left_layout .addWidget (self .filter_type_dropdown )
        left_layout .addWidget (self .cutoff_frequency_label )
        left_layout .addWidget (self .cutoff_frequency_spinbox )
        left_layout .addWidget (self .plot_btn )


        self .plot_scroll_area =QScrollArea ()
        self .plot_scroll_area .setWidgetResizable (True )
        self .plot_widget =QWidget ()
        self .plot_layout =QGridLayout ()
        self .plot_widget .setLayout (self .plot_layout )
        self .plot_scroll_area .setWidget (self .plot_widget )
        right_layout .addWidget (self .plot_scroll_area ,7 )


        self .plot_selection_dropdown =QComboBox ()
        self .plot_selection_dropdown .currentIndexChanged .connect (self .on_plot_selection_changed )
        right_layout .addWidget (self .plot_selection_dropdown )


        self .control_tabs =QTabWidget ()
        right_layout .addWidget (self .control_tabs ,3 )


        self .avg_tab =QWidget ()
        self .avg_layout =QVBoxLayout ()
        self .avg_tab .setLayout (self .avg_layout )
        self .calculate_avg_chk =QCheckBox ("Calculate and Plot Avg")
        self .avg_window_size_input =QLineEdit ("10")
        self .avg_layout .addWidget (self .calculate_avg_chk )
        self .avg_layout .addWidget (QLabel ("Avg Window Size (ms):"))
        self .avg_layout .addWidget (self .avg_window_size_input )
        self .control_tabs .addTab (self .avg_tab ,"Average")


        self .threshold_tab =QWidget ()
        self .threshold_layout =QVBoxLayout ()
        self .threshold_tab .setLayout (self .threshold_layout )
        self .threshold_chk =QCheckBox ("Select Regions within Threshold")
        self .threshold_value_input =QSpinBox ()
        self .threshold_value_input .setRange (0 ,1000000 )
        self .threshold_percentage_input =QSpinBox ()
        self .threshold_percentage_input .setRange (0 ,100 )
        self .threshold_percentage_input .setValue (75 )
        self .threshold_layout .addWidget (self .threshold_chk )
        self .threshold_layout .addWidget (QLabel ("Value for selecting regions (pA):"))
        self .threshold_layout .addWidget (self .threshold_value_input )
        self .threshold_layout .addWidget (QLabel ("Threshold (%):"))
        self .threshold_layout .addWidget (self .threshold_percentage_input )
        self .control_tabs .addTab (self .threshold_tab ,"Threshold")


        self .segments_tab =QWidget ()
        self .segments_layout =QVBoxLayout ()
        self .segments_tab .setLayout (self .segments_layout )
        self .select_segments_chk =QCheckBox ("Select Segments")
        self .segment_list =QListWidget ()
        self .add_segment_btn =QPushButton ("Add Segment")
        self .delete_segment_btn =QPushButton ("Delete Segment")
        self .segments_layout .addWidget (self .select_segments_chk )
        self .segments_layout .addWidget (self .segment_list )
        self .segments_layout .addWidget (self .add_segment_btn )
        self .segments_layout .addWidget (self .delete_segment_btn )
        self .control_tabs .addTab (self .segments_tab ,"Segments")



        self .global_settings_chk =QCheckBox ("Apply settings to all plots")
        right_layout .addWidget (self .global_settings_chk )


        self .save_btn =QPushButton ("Save Selected Data")
        self .save_btn .clicked .connect (self .save_selected_data )
        right_layout .addWidget (self .save_btn )


        self .calculate_avg_chk .stateChanged .connect (self .update_avg_settings )
        self .avg_window_size_input .editingFinished .connect (self .update_avg_settings )
        self .threshold_chk .stateChanged .connect (self .update_threshold_settings )
        self .threshold_value_input .valueChanged .connect (self .update_threshold_settings )
        self .threshold_percentage_input .valueChanged .connect (self .update_threshold_settings )
        self .select_segments_chk .stateChanged .connect (self .update_segment_settings )
        self .add_segment_btn .clicked .connect (self .add_segment )
        self .delete_segment_btn .clicked .connect (self .delete_segment )


        self .update_plot_btn =QPushButton ("Update Plot")
        self .update_plot_btn .clicked .connect (self .apply_settings_and_update_plots )
        right_layout .addWidget (self .update_plot_btn )

    def on_plot_selection_changed (self ,index ):
        if 0 <=index <self .plot_selection_dropdown .count ():
            self .current_file_path =self .plot_selection_dropdown .itemData (index )
            self .update_ui_from_settings ()

    def apply_settings_and_update_plots (self ):
        self .update_settings_from_ui ()
        if self .global_settings_chk .isChecked ():
            current_settings =self .plot_data_dict [self .current_file_path ].settings 
            for plot_data in self .plot_data_dict .values ():
                plot_data .settings =current_settings .copy ()

        for file_path ,plot_data in self .plot_data_dict .items ():
            if plot_data .settings !=plot_data .original_settings :
                self .update_single_plot (file_path )
                plot_data .original_settings =plot_data .settings .copy ()


    def toggle_low_pass_filter (self ,state ):
        self .filter_type_dropdown .setEnabled (state ==Qt .CheckState .Checked .value )
        self .cutoff_frequency_spinbox .setEnabled (state ==Qt .CheckState .Checked .value )

    def toggle_plot_multiple_files (self ,state ):

        self .files_list_widget .setSelectionMode (
        QListWidget .SelectionMode .MultiSelection if state ==Qt .CheckState .Checked .value 
        else QListWidget .SelectionMode .SingleSelection 
        )

    def select_folder (self ):
        options =QFileDialog .Option .ShowDirsOnly 
        directory =QFileDialog .getExistingDirectory (self ,"Select Folder","",options =options )
        if directory :
            self .folder_path_label .setText (f"Selected folder: {directory}")
            self .populate_file_list (directory ,self .include_subfolders_chk .isChecked ())

    def apply_low_pass_filter (self ,data ,cutoff_frequency ,type ,sampling_rate ):
        if type =='Butterworth':
            nyquist_rate =sampling_rate /2.0 
            cutoff =cutoff_frequency /nyquist_rate 
            sos =butter (N =8 ,Wn =cutoff ,btype ='low',analog =False ,output ='sos')
            zi =sosfilt_zi (sos )*data [0 ]
            filtered_data ,_ =sosfilt (sos ,data ,zi =zi )
            return filtered_data 
        elif type =='Bessel':
            nyquist_rate =sampling_rate /2.0 
            cutoff =cutoff_frequency /nyquist_rate 
            sos =bessel (N =8 ,Wn =cutoff ,btype ='low',analog =False ,output ='sos')
            zi =sosfilt_zi (sos )*data [0 ]
            filtered_data ,_ =sosfilt (sos ,data ,zi =zi )
            return filtered_data 
        elif type =='Chebyshev I':
            nyquist_rate =sampling_rate /2.0 
            cutoff =cutoff_frequency /nyquist_rate 
            sos =cheby1 (N =8 ,rp =0.1 ,Wn =cutoff ,btype ='low',analog =False ,output ='sos')
            zi =sosfilt_zi (sos )*data [0 ]
            filtered_data ,_ =sosfilt (sos ,data ,zi =zi )
            return filtered_data 
        elif type =='Chebyshev II':
            nyquist_rate =sampling_rate /2.0 
            cutoff =cutoff_frequency /nyquist_rate 
            sos =cheby2 (N =8 ,rs =40 ,Wn =cutoff ,btype ='low',analog =False ,output ='sos')
            zi =sosfilt_zi (sos )*data [0 ]
            filtered_data ,_ =sosfilt (sos ,data ,zi =zi )
            return filtered_data 
        elif type =='Elliptic':
            nyquist_rate =sampling_rate /2.0 
            cutoff =cutoff_frequency /nyquist_rate 
            sos =ellip (N =8 ,rp =0.1 ,rs =40 ,Wn =cutoff ,btype ='low',analog =False ,output ='sos')
            zi =sosfilt_zi (sos )*data [0 ]
            filtered_data ,_ =sosfilt (sos ,data ,zi =zi )
            return filtered_data 
        elif type =='FIR':
            nyquist_rate =sampling_rate /2.0 
            cutoff =cutoff_frequency /nyquist_rate 
            taps =firwin (101 ,cutoff )
            filtered_data =lfilter (taps ,1 ,data )
            return filtered_data 
        elif type =='IIR':
            nyquist_rate =sampling_rate /2.0 
            cutoff =cutoff_frequency /nyquist_rate 
            b ,a =butter (N =8 ,Wn =cutoff ,btype ='low',analog =False )
            filtered_data =lfilter (b ,a ,data )
            return filtered_data 

    def update_settings (self ):
        if self .global_settings_chk .isChecked ():
            settings =self .global_settings 
        elif self .current_file_path :
            settings =self .plot_data_dict [self .current_file_path ].settings 
        else :
            return 

        new_settings =Settings ()
        new_settings .calculate_avg =self .calculate_avg_chk .isChecked ()
        new_settings .avg_window_size =float (self .avg_window_size_input .text ())
        new_settings .use_threshold =self .threshold_chk .isChecked ()
        new_settings .threshold_value =self .threshold_value_input .value ()/1000 
        new_settings .threshold_percentage =self .threshold_percentage_input .value ()
        new_settings .use_segments =self .select_segments_chk .isChecked ()
        new_settings .segments =self .get_segments_from_list ()

        if self .global_settings_chk .isChecked ():
            self .global_settings =new_settings 
        else :
            self .plot_data_dict [self .current_file_path ].settings =new_settings 


    def update_ui_from_settings (self ):
        if not self .current_file_path :
            return 
        plot_data =self .plot_data_dict [self .current_file_path ]
        settings =plot_data .settings 


        self .calculate_avg_chk .setChecked (settings .get ('calculate_avg',False ))
        self .avg_window_size_input .setText (str (settings .get ('avg_window_size',10 )))
        self .threshold_chk .setChecked (settings .get ('use_threshold',False ))
        self .threshold_value_input .setValue (int (settings .get ('threshold_value',0 )*1000 ))
        self .threshold_percentage_input .setValue (settings .get ('threshold_percentage',75 ))


        self .segment_list .clear ()


        if 'segments'in settings and settings ['segments']:
            for start ,end in settings ['segments']:
                self .segment_list .addItem (f"{start:.2f} - {end:.2f}")
            self .select_segments_chk .setChecked (True )
        else :
            self .select_segments_chk .setChecked (False )


        settings ['use_segments']=self .select_segments_chk .isChecked ()



















    def update_settings_from_ui (self ):
        if not self .current_file_path :
            return 
        plot_data =self .plot_data_dict [self .current_file_path ]
        settings =plot_data .settings 
        settings ['calculate_avg']=self .calculate_avg_chk .isChecked ()
        settings ['avg_window_size']=float (self .avg_window_size_input .text ())
        settings ['use_threshold']=self .threshold_chk .isChecked ()
        settings ['threshold_value']=self .threshold_value_input .value ()/1000 
        settings ['threshold_percentage']=self .threshold_percentage_input .value ()
        settings ['use_segments']=self .select_segments_chk .isChecked ()
        settings ['segments']=self .get_segments_from_list ()

    def populate_file_list (self ,directory ,include_subfolders ):
        self .files_list_widget .clear ()
        self .plot_data_dict .clear ()
        file_list =[]

        for root ,dirs ,files in os .walk (directory ):
            for file in files :
                if file .endswith ('.abf')or file .endswith ('.hdf5')or file .endswith ('.h5'):
                    rel_path =os .path .relpath (os .path .join (root ,file ),start =directory )
                    full_path =os .path .join (root ,file )
                    file_list .append ((rel_path ,full_path ))

            if not include_subfolders :
                break 

        file_list .sort (key =lambda x :x [0 ].lower ())

        for rel_path ,full_path in file_list :
            item =QListWidgetItem (rel_path )
            item .setData (Qt .ItemDataRole .UserRole ,full_path )
            self .files_list_widget .addItem (item )

    def plot_selected_files (self ):
        selected_items =self .files_list_widget .selectedItems ()
        if not selected_items :
            return 

        self .plot_data_dict .clear ()
        self .clear_plot_layout ()

        if not self .plot_multiple_files_chk .isChecked ():
            selected_items =[selected_items [0 ]]

        total_steps =len (selected_items )*2 
        progress =QProgressDialog ("Processing files...","Cancel",0 ,total_steps ,self )
        progress .setWindowModality (Qt .WindowModal )
        progress .setWindowTitle ("Please Wait")
        progress .setMinimumDuration (0 )

        current_step =0 


        for item in selected_items :
            if progress .wasCanceled ():
                return 
            file_path =item .data (Qt .ItemDataRole .UserRole )
            progress .setLabelText (f"Loading {os.path.basename(file_path)}...")
            try :
                data ,unfiltered_data ,sampling_rate ,time =self .load_file (file_path )
                plot_data =PlotData (file_path ,data ,unfiltered_data ,sampling_rate ,time )
                self .plot_data_dict [file_path ]=plot_data 
            except Exception as e :
                QMessageBox .warning (self ,"Error",str (e ))
            current_step +=1 
            progress .setValue (current_step )
            QApplication .processEvents ()


        num_plots =len (self .plot_data_dict )
        cols =math .ceil (math .sqrt (num_plots ))
        rows =math .ceil (num_plots /cols )

        for i ,(file_path ,plot_data )in enumerate (self .plot_data_dict .items ()):
            if progress .wasCanceled ():
                break 
            progress .setLabelText (f"Plotting {os.path.basename(file_path)}...")

            row =i //cols 
            col =i %cols 


            plot_widget =QWidget ()
            plot_layout =QVBoxLayout (plot_widget )
            plot_layout .setSpacing (0 )
            plot_layout .setContentsMargins (0 ,0 ,0 ,0 )

            figure =Figure (figsize =(5 ,3 ),dpi =100 )
            canvas =FigureCanvasQTAgg (figure )
            ax =figure .add_subplot (111 )

            self .plot_data (ax ,plot_data )


            toolbar =NavigationToolbar2QT (canvas ,plot_widget )
            toolbar .setIconSize (QSize (16 ,16 ))
            toolbar .setStyleSheet ("""
                QToolBar { border: 0px }
                QToolButton { border: 0px; padding: 1px; }
            """)

            plot_layout .addWidget (canvas )
            plot_layout .addWidget (toolbar )

            self .plot_layout .addWidget (plot_widget ,row ,col )

            current_step +=1 
            progress .setValue (current_step )
            QApplication .processEvents ()

        progress .close ()


        self .plot_selection_dropdown .clear ()
        for file_path in self .plot_data_dict .keys ():
            self .plot_selection_dropdown .addItem (os .path .basename (file_path ),file_path )

        if self .plot_data_dict :
            self .current_file_path =next (iter (self .plot_data_dict ))
            self .update_settings_display (self .plot_data_dict [self .current_file_path ])

    def update_selected_plots (self ):
        if self .global_settings_chk .isChecked ():
            for file_path ,plot_data in self .plot_data_dict .items ():
                self .update_single_plot (file_path )
        elif self .current_file_path :
            self .update_single_plot (self .current_file_path )


    def update_single_plot (self ,file_path ):
        plot_data =self .plot_data_dict [file_path ]
        index =list (self .plot_data_dict .keys ()).index (file_path )
        plot_widget =self .plot_layout .itemAt (index ).widget ()
        canvas =plot_widget .layout ().itemAt (0 ).widget ()
        ax =canvas .figure .axes [0 ]
        ax .clear ()

        self .plot_data (ax ,plot_data )

        if plot_data .settings ['calculate_avg']and plot_data .settings ['avg_window_size']>0 :
            rolling_avg =calculate_rolling_stats (plot_data .data ,plot_data .sampling_rate ,plot_data .settings ['avg_window_size'])
            ax .plot (plot_data .time ,rolling_avg ,linewidth =2 ,color ='red')

        if plot_data .settings ['use_segments']:
            self .plot_segments (ax ,plot_data )
        elif plot_data .settings ['use_threshold']:
            self .plot_threshold (ax ,plot_data )

        canvas .draw ()
        plot_data .original_settings =plot_data .settings .copy ()


    def update_settings_display (self ,plot_data ):
        settings =plot_data .settings 
        self .calculate_avg_chk .setChecked (settings ['calculate_avg'])
        self .avg_window_size_input .setText (str (settings ['avg_window_size']))
        self .threshold_chk .setChecked (settings ['use_threshold'])
        self .threshold_value_input .setValue (int (settings ['threshold_value']*1000 ))
        self .threshold_percentage_input .setValue (settings ['threshold_percentage'])
        self .select_segments_chk .setChecked (settings ['use_segments'])
        self .segment_list .clear ()
        for start ,end in settings ['segments']:
            self .segment_list .addItem (f"{start:.2f} - {end:.2f}")


    def update_avg_settings (self ):
        self .update_settings_from_ui ()

    def update_threshold_settings (self ):
        self .update_settings_from_ui ()

    def update_segment_settings (self ):
        self .update_settings_from_ui ()


    def plot_data (self ,ax ,plot_data ):
        nth_element =self .nth_element_spinbox .value ()
        ax .plot (plot_data .time [::nth_element ],plot_data .data [::nth_element ],linewidth =0.5 )
        ax .set_xlabel ('Time (s)')
        ax .set_ylabel ('Current (nA)')
        ax .set_xlim ([min (plot_data .time ),max (plot_data .time )])
        ax .set_title (os .path .basename (plot_data .file_path ))


    def add_plot_tab (self ,title ):
        tab =QWidget ()
        tab_layout =QVBoxLayout ()
        tab .setLayout (tab_layout )

        figure =Figure (figsize =(5 ,4 ),dpi =100 )
        canvas =FigureCanvasQTAgg (figure )
        tab_layout .addWidget (canvas )

        mpl_toolbar =NavigationToolbar2QT (canvas ,self )
        tab_layout .addWidget (mpl_toolbar )

        self .right_tab_widget .addTab (tab ,title )
        return figure ,canvas 


    def load_file (self ,file_path ):
        try :
            if file_path .endswith ('.abf'):
                raw_data ,sampling_rate ,time =load_abf_file (file_path )
            elif file_path .endswith ('.hdf5')or file_path .endswith ('.h5'):
                raw_data ,sampling_rate ,time =load_hdf5_file (file_path )
            else :
                raise ValueError (f"Unsupported file format: {file_path}")

            unfiltered_data =raw_data .copy ()
            data =raw_data .copy ()

            if self .low_pass_filter_chk .isChecked ():
                filter_type =self .filter_type_dropdown .currentText ()
                cutoff_frequency =self .cutoff_frequency_spinbox .value ()*1000 

                if sampling_rate <=cutoff_frequency *2 :
                    raise ValueError (f"The selected cutoff frequency is too high for the sampling rate of {file_path}.")

                data =self .apply_low_pass_filter (data ,cutoff_frequency ,filter_type ,sampling_rate )

            return data ,unfiltered_data ,sampling_rate ,time 
        except Exception as e :
            raise Exception (f"Error loading {os.path.basename(file_path)}: {str(e)}")

    def clear_plot_layout (self ):
        while self .plot_layout .count ():
            item =self .plot_layout .takeAt (0 )
            widget =item .widget ()
            if widget :
                widget .deleteLater ()

        self .plot_layout .setSpacing (5 )
        self .plot_layout .setContentsMargins (5 ,5 ,5 ,5 )

    def find_contiguous_regions (self ,data ):
        contiguous_regions =[]
        data =np .sort (data )
        start_index =data [0 ]
        prev_index =start_index 
        for i in range (1 ,len (data )):
            if data [i ]>prev_index +1 :
                contiguous_regions .append ((start_index ,prev_index ))
                start_index =data [i ]
            prev_index =data [i ]
        contiguous_regions .append ((start_index ,data [-1 ]))
        return contiguous_regions 

    def plot_threshold (self ,ax ,plot_data ):
        settings =plot_data .settings 
        threshold_value =settings ['threshold_value']
        threshold_percentage =settings ['threshold_percentage']/100 
        avg_window_size_in_ms =settings ['avg_window_size']

        if avg_window_size_in_ms >0 :
            rolling_avg =calculate_rolling_stats (plot_data .data ,plot_data .sampling_rate ,avg_window_size_in_ms )
        else :
            rolling_avg =plot_data .data 

        threshold_lower =threshold_value -(threshold_value *(1 -threshold_percentage ))
        threshold_upper =threshold_value +(threshold_value *(1 -threshold_percentage ))

        selected_regions =np .where ((rolling_avg >=threshold_lower )&(rolling_avg <=threshold_upper ))[0 ]

        for region_start ,region_end in self .find_contiguous_regions (selected_regions ):
            start_time =plot_data .time [region_start ]
            end_time =plot_data .time [region_end ]
            ax .axvline (x =start_time ,color ='r',linestyle ='-',linewidth =0.5 )
            ax .axvline (x =end_time ,color ='r',linestyle ='-',linewidth =0.5 )
            ax .axvspan (start_time ,end_time ,alpha =0.3 ,color ='red')


    def plot_segments (self ,ax ,plot_data ):
        for start_time ,end_time in plot_data .settings ['segments']:
            ax .axvline (x =start_time ,color ='g',linestyle ='-',linewidth =0.5 )
            ax .axvline (x =end_time ,color ='g',linestyle ='-',linewidth =0.5 )
            ax .axvspan (start_time ,end_time ,alpha =0.3 ,color ='green')

    def toggle_select_segments (self ,state ):
        if state ==Qt .CheckState .Checked .value :
            self .segment_list .setEnabled (True )
            self .add_segment_btn .setEnabled (True )
            self .delete_segment_btn .setEnabled (True )
        else :
            self .segment_list .setEnabled (False )
            self .add_segment_btn .setEnabled (False )
            self .delete_segment_btn .setEnabled (False )
            self .segment_list .clear ()

        self .update_settings ()
        self .update_selected_plots ()

    def add_segment (self ):
        start_time ,ok =QInputDialog .getDouble (self ,"Add Segment","Start time (s):")
        if ok :
            end_time ,ok =QInputDialog .getDouble (self ,"Add Segment","End time (s):")
            if ok :
                self .segment_list .addItem (f"{start_time:.2f} - {end_time:.2f}")
                self .update_settings_from_ui ()

    def delete_segment (self ):
        current_row =self .segment_list .currentRow ()
        if current_row >=0 :
            self .segment_list .takeItem (current_row )
            self .update_settings_from_ui ()

    def get_segments_from_list (self ):
        segments =[]
        for i in range (self .segment_list .count ()):
            item =self .segment_list .item (i )
            start ,end =map (float ,item .text ().split (' - '))
            segments .append ((start ,end ))
        return segments 

    def save_selected_data (self ):
        if not self .plot_data_dict :
            QMessageBox .warning (self ,"Warning","No data to save.")
            return 
        directory =QFileDialog .getExistingDirectory (self ,"Select Directory to Save Files")
        if not directory :
            return 
        progress =QProgressDialog ("Saving files...","Cancel",0 ,len (self .plot_data_dict ),self )
        progress .setWindowModality (Qt .WindowModal )
        progress .setWindowTitle ("Please Wait")
        progress .setMinimumDuration (0 )

        saved_files =[]
        for i ,(file_path ,plot_data )in enumerate (self .plot_data_dict .items ()):
            if progress .wasCanceled ():
                break 
            try :
                result =self .save_plot_data (plot_data ,directory )
                if result :
                    saved_files .append (os .path .basename (result ))
                else :
                    QMessageBox .warning (self ,"Warning",f"No data selected for {os.path.basename(file_path)}. File not saved.")
            except Exception as e :
                QMessageBox .warning (self ,"Error",f"Failed to save data for {os.path.basename(file_path)}: {str(e)}")
            progress .setValue (i +1 )
            QApplication .processEvents ()
        progress .close ()

        if saved_files :
            message =f"Data saving process completed. {len(saved_files)} file(s) saved:\n"+"\n".join (saved_files )
            QMessageBox .information (self ,"Success",message )
        else :
            QMessageBox .warning (self ,"Warning","No files were saved.")

    def save_plot_data (self ,plot_data ,directory ):
        file_name =os .path .basename (plot_data .file_path )
        unfiltered_data =plot_data .unfiltered_data 
        sampling_rate =plot_data .sampling_rate 
        time =plot_data .time 
        settings =plot_data .settings 
        selected_indices =self .get_selected_indices (plot_data )

        if not np .any (selected_indices ):
            return False 

        selected_data =unfiltered_data [selected_indices ]
        selected_time =time [selected_indices ]

        base_name ,ext =os .path .splitext (file_name )
        output_filename =f"{base_name}_selected_unfiltered_data.h5"
        output_path =os .path .join (directory ,output_filename )

        with h5py .File (output_path ,'w')as f :
            f .create_dataset ('selected_data',data =selected_data )
            f .attrs ['sampling_rate']=sampling_rate 
            f .attrs ['original_file']=file_name 
            f .attrs ['duration']=selected_time [-1 ]-selected_time [0 ]

        return output_path 

    def get_selected_indices (self ,plot_data ):
        settings =plot_data .settings 
        time =plot_data .time 
        if settings ['use_threshold']:

            rolling_avg =calculate_rolling_stats (plot_data .data ,plot_data .sampling_rate ,settings ['avg_window_size'])
            threshold_value =settings ['threshold_value']
            threshold_percentage =settings ['threshold_percentage']/100 
            threshold_lower =threshold_value -(threshold_value *(1 -threshold_percentage ))
            threshold_upper =threshold_value +(threshold_value *(1 -threshold_percentage ))
            return np .where ((rolling_avg >=threshold_lower )&(rolling_avg <=threshold_upper ))[0 ]
        elif settings ['use_segments']:
            indices =[]
            sorted_segments =sorted (settings ['segments'],key =lambda x :x [0 ])
            for start ,end in sorted_segments :
                start_idx =np .searchsorted (time ,start )
                end_idx =np .searchsorted (time ,end )
                indices .extend (range (start_idx ,end_idx +1 ))
            return np .array (indices )
        else :
            return np .arange (len (plot_data .data ))



if __name__ =='__main__':
    app =QApplication (sys .argv )
    app .setStyle (QStyleFactory .create ('Fusion'))

    palette =QPalette ()
    palette .setColor (QPalette .ColorRole .Window ,QColor (53 ,53 ,53 ))
    palette .setColor (QPalette .ColorRole .WindowText ,Qt .GlobalColor .white )
    palette .setColor (QPalette .ColorRole .Base ,QColor (25 ,25 ,25 ))
    palette .setColor (QPalette .ColorRole .AlternateBase ,QColor (53 ,53 ,53 ))
    palette .setColor (QPalette .ColorRole .ToolTipBase ,Qt .GlobalColor .white )
    palette .setColor (QPalette .ColorRole .ToolTipText ,Qt .GlobalColor .white )
    palette .setColor (QPalette .ColorRole .Text ,Qt .GlobalColor .white )
    palette .setColor (QPalette .ColorRole .Button ,QColor (53 ,53 ,53 ))
    palette .setColor (QPalette .ColorRole .ButtonText ,Qt .GlobalColor .white )
    palette .setColor (QPalette .ColorRole .BrightText ,Qt .GlobalColor .red )
    palette .setColor (QPalette .ColorRole .Link ,QColor (42 ,130 ,218 ))
    palette .setColor (QPalette .ColorRole .Highlight ,QColor (42 ,130 ,218 ))
    palette .setColor (QPalette .ColorRole .HighlightedText ,Qt .GlobalColor .black )
    app .setPalette (palette )
    window =MainWindow ()
    window .showMaximized ()
    sys .exit (app .exec ())